#!/usr/bin/env python
# -*-coding:utf-8 -*-

import os
import torch
import optuna
import shutil
import joblib
import numpy as np
import pandas as pd
import matplotlib
matplotlib.use('Agg')
import matplotlib.pyplot as plt
import seaborn as sns
from tqdm import tqdm
from copy import copy
from sklearn.preprocessing import StandardScaler
from sklearn.model_selection import KFold


def get_std(data, scaler):
    data_std = scaler.transform(data)
    return data_std


def get_scaler(data_scaler):
    data = pd.read_csv(data_scaler)
    scaler = StandardScaler()
    scaler.fit(data.to_numpy())
    return scaler


def define_model(trial):
    n_layers = trial.suggest_int('n_layers', LAYERS_MIN, LAYERS_MAX, step=LAYERS_STEP)
    layers = []
    input_features = pd.read_csv(INPUT_DATA).shape[1]
    output_features = pd.read_csv(OUTPUT_DATA).shape[1]
    in_features = copy(input_features)

    for i in range(n_layers):
        out_features = trial.suggest_int(f'n_units_l{i}', NODES_MIN, NODES_MAX, step=NODES_STEP)
        layers.append(torch.nn.Linear(in_features, out_features))
        layers.append(torch.nn.Sigmoid())
        p = trial.suggest_float("dropout_l{}".format(i), DROPOUT_MIN, DROPOUT_MAX)
        layers.append(torch.nn.Dropout(p))
        in_features = out_features
    layers.append(torch.nn.Linear(in_features, output_features))
    return torch.nn.Sequential(*layers)


def objective(trial):
    # Get the training and validation data
    x_scaler = get_scaler(data_scaler=INPUT_DATA)
    y_scaler = get_scaler(data_scaler=OUTPUT_DATA)
    
    # Perform K-Fold cross validation
    input_data = pd.read_csv(INPUT_DATA)
    output_data = pd.read_csv(OUTPUT_DATA)
    x = input_data.to_numpy()
    y = output_data.to_numpy()
    x_std = get_std(x, x_scaler)
    y_std = get_std(y, y_scaler)

    # Training with K-Fold cross validation
    loss_train_data, loss_valid_data = [], []
    loss_train_fold, loss_valid_fold = [], []
    for fold, (train_index, valid_index) in enumerate(KFold(n_splits=5, shuffle=True, random_state=0).split(x_std)):
        # Define the model
        torch.manual_seed(0)
        model = define_model(trial).to(DEVICE)
        optimizer_name = trial.suggest_categorical('optimizer', OPTIMIZERS)
        lr = trial.suggest_float('lr', LR_MIN, LR_MAX, log=True)
        wd = trial.suggest_float('weight_decay', WD_MIN, WD_MAX, log=True)
        optimizer = getattr(torch.optim, optimizer_name)(model.parameters(), lr=lr, weight_decay=wd)
        loss_func = torch.nn.MSELoss()

        # Split the data into training and validation sets
        x_train, x_valid = x_std[train_index], x_std[valid_index]
        y_train, y_valid = y_std[train_index], y_std[valid_index]
        data_train, data_valid = [x_train, y_train], [x_valid, y_valid]

        # Transfer the data as torch tensors
        train_x, train_y = torch.FloatTensor(data_train[0]).to(DEVICE), torch.FloatTensor(data_train[1]).to(DEVICE)
        valid_x, valid_y = torch.FloatTensor(data_valid[0]).to(DEVICE), torch.FloatTensor(data_valid[1]).to(DEVICE)

        # Dynamically change the number of epochs based on the NN complexity
        layers = (model.__len__() - 1) // 3
        epochs = 100 + 10 * layers

        # Training
        model.train()
        for epoch in tqdm(range(epochs), desc=f'Trial {trial.number} Fold {fold+1}', position=0, leave=False):
            optimizer.zero_grad()
            output_train = model(train_x)
            output_valid = model(valid_x)
            loss_train = loss_func(output_train, train_y)
            loss_valid = loss_func(output_valid, valid_y)
            loss_train_data.append([fold, epoch, loss_train.item()])
            loss_valid_data.append([fold, epoch, loss_valid.item()])
            loss_train.backward()
            optimizer.step()

        # Validation
        with torch.no_grad():
            model.eval()
            train_y_pred = model(train_x)
            valid_y_pred = model(valid_x)
            loss_train_pred = loss_func(train_y_pred, train_y)
            loss_valid_pred = loss_func(valid_y_pred, valid_y)
            loss_train_fold.append([fold, loss_train_pred.item()])
            loss_valid_fold.append([fold, loss_valid_pred.item()])

    # Get the average validation loss
    loss_train_fold = np.array(loss_train_fold)
    loss_valid_fold = np.array(loss_valid_fold)
    loss_train_avg = loss_train_fold[:, 1].mean()
    loss_valid_avg = loss_valid_fold[:, 1].mean()

    # Define the accuracy
    accuracy = loss_valid_avg

    global CURRENT_ACCURACY, FOUND_NEW
    if accuracy < CURRENT_ACCURACY:
        CURRENT_ACCURACY = accuracy
        FOUND_NEW = True
    else:
        FOUND_NEW = False

    # Plot the loss curve
    if FOUND_NEW:
        loss_train_data = np.array(loss_train_data)
        loss_valid_data = np.array(loss_valid_data)
        sns.set_theme(font_scale=1.2, style='whitegrid')
        matplotlib.rcParams['xtick.direction'] = 'in'
        matplotlib.rcParams['ytick.direction'] = 'in'
        fig = plt.figure(figsize=(13, 6), constrained_layout=True)
        for fold in range(6):
            if fold != 5:
                # Plot the loss curve for each fold
                ax = plt.subplot(3, 2, fold+1)
                fold_train = loss_train_data[loss_train_data[:, 0] == fold]
                fold_valid = loss_valid_data[loss_valid_data[:, 0] == fold]
                ax.plot(fold_train[:, 1], fold_train[:, 2], label='Training', color='C0', linewidth=1.5)
                ax.plot(fold_valid[:, 1], fold_valid[:, 2], label='Validation', color='C1', linewidth=1.5)
                ax.text(0.05, 0.95, f'Fold {fold+1}', transform=ax.transAxes, fontweight='bold', va='top', ha='left')
                ax.set_xlabel('Epoch')
                ax.set_ylabel('Loss')
                ax.legend(loc='upper right', frameon=True, ncol=2)
            elif fold == 5:
                # Plot the average loss
                ax = plt.subplot(3, 2, fold+1)
                ax.plot(loss_train_fold[:, 0], loss_train_fold[:, 1], label='Training', color='C0', linewidth=0, marker='o', markersize=10, markerfacecolor='none', markeredgecolor='C0', markeredgewidth=1.5)
                ax.plot(loss_valid_fold[:, 0], loss_valid_fold[:, 1], label='Validation', color='C1', linewidth=0, marker='o', markersize=10, markerfacecolor='none', markeredgecolor='C1', markeredgewidth=1.5)
                ax.axhline(loss_train_avg, color='C0', linestyle='--', linewidth=2.0, label=f'Avg. Training = {loss_train_avg:.2e}')
                ax.axhline(loss_valid_avg, color='C1', linestyle='--', linewidth=2.0, label=f'Avg. Validation = {loss_valid_avg:.2e}')
                ax.set_xticks(range(5))
                ax.set_xticklabels(range(1, 6))
                ax.set_xlabel('Fold')
                ax.legend(loc='upper right', frameon=True, ncol=2)
                ax.set_ylabel('Loss')
        plt.savefig(f'{SAVE_PATH}/best_model_losses/loss_model_{trial.number}.png', dpi=330)
        plt.close()

    # Handle pruning based on the  intermediate value.
    trial.report(accuracy, epoch)
    if trial.should_prune():
        raise optuna.exceptions.TrialPruned()
    
    # Save the model
    torch.save(model, f'{SAVE_PATH}/tested_models/model_{trial.number}.pkl')

    return accuracy


def optimize(input_data, output_data, n_trials, save_path):
    global INPUT_DATA, OUTPUT_DATA, N_TRIALS, SAVE_PATH, FOUND_NEW, CURRENT_ACCURACY, LAYERS_MIN, LAYERS_MAX, LAYERS_STEP, NODES_MIN, NODES_MAX, NODES_STEP, OPTIMIZERS, LR_MIN, LR_MAX, WD_MIN, WD_MAX, DROPOUT_MIN, DROPOUT_MAX, DEVICE

    INPUT_DATA = input_data
    OUTPUT_DATA = output_data
    N_TRIALS = n_trials
    SAVE_PATH = save_path
    FOUND_NEW = False
    CURRENT_ACCURACY = 1e10
    LAYERS_MIN, LAYERS_MAX, LAYERS_STEP = 2, 10, 1
    NODES_MIN, NODES_MAX, NODES_STEP = 8, 256, 8
    OPTIMIZERS= ['Adam', 'SGD', 'RMSprop']
    LR_MIN, LR_MAX = 1e-4, 1e-2
    WD_MIN, WD_MAX = 1e-6, 1e-2
    DROPOUT_MIN, DROPOUT_MAX = 0.0, 0.5
    DEVICE = torch.device('cuda' if torch.cuda.is_available() else 'cpu')

    # Create directory for storing results
    if not os.path.exists(SAVE_PATH):
        os.makedirs(SAVE_PATH)
    if not os.path.exists(f'{SAVE_PATH}/tested_models'):
        os.makedirs(f'{SAVE_PATH}/tested_models')
    if not os.path.exists(f'{SAVE_PATH}/best_model_losses'):
        os.makedirs(f'{SAVE_PATH}/best_model_losses')

    # Set random seed for reproducibility
    torch.manual_seed(42)
    np.random.seed(42)

    # Optuna process
    study = optuna.create_study(direction='minimize', sampler=optuna.samplers.TPESampler(seed=42))
    study.optimize(objective, n_trials=N_TRIALS, timeout=None)

    # Optimization complete
    pruned_trials = study.get_trials(deepcopy=False, states=[optuna.trial.TrialState.PRUNED])
    complete_trials = study.get_trials(deepcopy=False, states=[optuna.trial.TrialState.COMPLETE])
    best_trial = study.best_trial

    # Write log
    with open(f'{SAVE_PATH}/best_model_params.log', 'a') as f:
        f.write(f'Study statistics:\n')
        f.write(f'  Number of finished trials: {len(study.trials)}\n')
        f.write(f'  Number of pruned trials: {len(pruned_trials)}\n')
        f.write(f'  Number of complete trials: {len(complete_trials)}\n')
        f.write(f'Best trial:\n')
        f.write(f'  Number: {best_trial.number}\n')
        f.write(f'  Value: {best_trial.value}\n')
        f.write(f'  Params:\n')
        for key, value in best_trial.params.items():
            f.write(f'    {key}: {value}\n')
    
    # Select target model
    old_file_path = f'{SAVE_PATH}/tested_models/model_{best_trial.number}.pkl'
    new_file_path = f'{SAVE_PATH}/best_model.pkl'
    shutil.copyfile(old_file_path, new_file_path)
    
    # Save study object
    joblib.dump(study, f'{SAVE_PATH}/study.pkl')
    
    # # Remove tested models
    # shutil.rmtree(f'{SAVE_PATH}/tested_models')

    # Draw the study results
    study_path = f'{SAVE_PATH}/study.pkl'
    draw_study(study_path)


def draw_study(study_path):
    study = joblib.load(study_path)
    best_trial_number = study.best_trial.number
    best_trial_value = study.best_trial.value

    # Save study results to csv
    data = study.trials_dataframe()
    pd.DataFrame(data).to_csv(f'{study_path[:-4]}.csv', index=False)

    # Get the complete trials data
    data_complete = data[data['state'] == 'COMPLETE']
    number = data_complete.loc[:, 'number']
    value = data_complete.loc[:, 'value']
    lr = data_complete.loc[:, 'params_lr']
    n_layers = data_complete.loc[:, 'params_n_layers']
    optimizer = data_complete.loc[:, 'params_optimizer']
    weight_decay = data_complete.loc[:, 'params_weight_decay']
    value_min = [value.iloc[:i+1].min() for i in range(len(value))]

    # Repeat the last value_min
    value_min = value_min + [value_min[-1]]

    # Add the last number to number_min
    number_min = number.copy().to_list()
    number_min = number_min + [len(data)]

    # Get the pruned trials data
    data_pruned = data[data['state'] == 'PRUNED']
    pruned_number = data_pruned.loc[:, 'number']
    pruned_value = data_pruned.loc[:, 'value']

    sns.set_theme(font_scale=1.2, style='whitegrid')
    matplotlib.rcParams['xtick.direction'] = 'in'
    matplotlib.rcParams['ytick.direction'] = 'in'

    # History
    fig = plt.figure(figsize=(13, 4), constrained_layout=True)
    ax = plt.subplot()
    ax.scatter(number, value, label='Completed Trials', marker='o', facecolor='none', edgecolors='C0', linewidths=1.0, s=20)
    ax.scatter(pruned_number, pruned_value, label='Pruned Trials', marker='o', facecolor='none', edgecolors='C7', linewidths=1.0, s=20, alpha=0.5)
    ax.plot(number_min, value_min, c='r', label='Best Value')
    ax.scatter(best_trial_number, best_trial_value, marker='o', facecolor='C3', edgecolors='C3', linewidths=1.5, s=20)
    ax.scatter(best_trial_number, best_trial_value, marker='*', facecolor='C3', edgecolors='C3', linewidths=1.5, s=100, label=f'Best Trial: #{best_trial_number}')
    ax.set_xlabel('Number of Trials')
    ax.set_ylabel('Objective Value')
    ax.legend(loc='upper right', frameon=True)
    plt.savefig(f'{study_path[:-4]}_history.png', dpi=330)
    plt.close()
    # print(f'Plot of study history saved to {study_path[:-4]}_history.png')

    # Importance
    x = ['weight_decay', 'optimizer', 'lr', 'n_layers']
    x_labels = ['Weight Decay', 'Optimizer', 'Learning Rate', 'Layers']
    importance = optuna.importance.get_param_importances(study, params=x)
    h = [importance[x[0]], importance[x[1]], importance[x[2]], importance[x[3]]]
    fig = plt.figure(figsize=(4, 4), constrained_layout=True)
    ax = plt.subplot()
    ax_twin = ax.twiny()
    ax.set_zorder(ax_twin.get_zorder() + 1)
    ax.bar(x_labels, h, color='C0', edgecolor='black', linewidth=1.0)
    ax.set_ylim(top=max(h) * 1.1)
    ax_twin.set_xlim(ax.get_xlim())
    ax_twin.set_xticks(range(len(x_labels)))
    ax_twin.set_xticklabels(x_labels, rotation=15)
    ax.set_xticks(x_labels)
    ax.set_xticklabels([])
    ax.set_xlabel('Hyperparameters')
    ax.set_ylabel('Importance for Objective Value')
    plt.savefig(f'{study_path[:-4]}_importance.png', dpi=330)
    plt.close()
    # print(f'Plot of study importance saved to {study_path[:-4]}_importance.png')

    # Slice
    fig, ax = plt.subplots(1, 4, figsize=(9, 4), constrained_layout=True)
    a = ax[0].scatter(weight_decay, value, c=number, cmap='Blues', edgecolor='gray', linewidths=0.1)
    ax[0].set_xscale('log')
    ax[0].set_title('Weight Decay')
    ax[0].set_ylabel('Objective Value')
    ax[1].scatter(optimizer, value, c=number, cmap='Blues', edgecolor='gray', linewidths=0.1)
    ax[1].set_title('Optimizer')
    ax[1].set_xticks(range(len(optimizer.unique())))
    ax[1].set_xticklabels(optimizer.unique())
    ax[1].set_xlim(-0.5, len(optimizer.unique()) - 0.5)
    ax[2].scatter(lr, value, c=number, cmap='Blues', edgecolor='gray', linewidths=0.1)
    ax[2].set_xscale('log')
    ax[2].set_title('Learning Rate')
    ax[3].scatter(n_layers, value, c=number, cmap='Blues', edgecolor='gray', linewidths=0.1)
    ax[3].set_xticks(n_layers)
    ax[3].set_xticklabels(n_layers)
    ax[3].set_title('Layers')
    ax[3].set_xlim(min(n_layers) - 1, max(n_layers) + 1)
    cb = plt.colorbar(a, ax=ax, pad=0.015, aspect=40)
    cb.ax.set_title('# Trials')
    plt.savefig(f'{study_path[:-4]}_slice.png', dpi=330)
    plt.close()
    # print(f'Plot of study slice saved to {study_path[:-4]}_slice.png')


if __name__ == '__main__':
    optimize(
        input_data=INPUT_DATA,
        output_data=OUTPUT_DATA,
        n_trials=N_TRIALS,
        save_path=SAVE_PATH,
    )


#!/usr/bin/env python3

import sys, os, time
from bullet import Bullet, colors
from yaspin import yaspin
from yaspin.spinners import Spinners

from masgent import tools, schemas
from masgent.utils.utils import (
    color_print, 
    color_input, 
    print_help, 
    global_commands, 
    start_new_session,
    clear_and_print_entry_message,
    exit_and_cleanup,
    )

COMMANDS = {}

def register(code, func):
    def decorator(func):
        COMMANDS[code] = {
            'function': func,
            'description': func.__doc__ or ''
        }
        return func
    return decorator

def run_command(code):
    cmd = COMMANDS.get(code)
    if cmd:
        cmd['function']()
    else:
        color_print(f'[Error] Invalid command code: {code}\n', 'red')

def check_poscar():
    while True:
        runs_dir = os.environ.get('MASGENT_SESSION_RUNS_DIR')

        if os.path.exists(os.path.join(runs_dir, 'POSCAR')):
            use_default = True
        else:
            use_default = False

        if use_default:
            runs_dir_name = os.path.basename(runs_dir)
            clear_and_print_entry_message()
            choices = [
                'Yes  ->  Use POSCAR file in current runs directory',
                'No   ->  Provide a different POSCAR file path',
            ] + global_commands()

            prompt = f'\nUse POSCAR file in current runs directory: {runs_dir_name}/POSCAR ?\n'
            cli = Bullet(prompt=prompt, choices=choices, margin=1, bullet=' ●', word_color=colors.foreground['green'])
            user_input = cli.launch()

            if user_input.startswith('AI'):
                from masgent.ai_mode import ai_backend
                ai_backend.main()
            elif user_input.startswith('New'):
                start_new_session()
            elif user_input.startswith('Back'):
                return
            elif user_input.startswith('Main'):
                run_command('0')
            elif user_input.startswith('Help'):
                print_help()
            elif user_input.startswith('Exit'):
                exit_and_cleanup()
            elif user_input.startswith('Yes'):
                poscar_path = os.path.join(runs_dir, 'POSCAR')
            elif user_input.startswith('No'):
                poscar_path = color_input('\nEnter path to POSCAR file: ', 'yellow').strip()
            else:
                continue
        else:
            poscar_path = color_input('\nEnter path to POSCAR file: ', 'yellow').strip()
        
        if not poscar_path:
            continue
        
        try:
            schemas.CheckPoscar(poscar_path=poscar_path)
            return poscar_path
        except Exception:
            color_print(f'[Error] Invalid POSCAR: {poscar_path}, please double check and try again.\n', 'red')

#############################################
#                                           #
# Below are implementations of sub-commands #
#                                           #
#############################################

@register('1.1.1', 'Generate POSCAR from chemical formula.')
def command_1_1_1():
    try: 
        while True:
            formula = color_input('\nEnter chemical formula (e.g., NaCl): ', 'yellow').strip()

            if not formula:
                continue

            try:
                schemas.GenerateVaspPoscarSchema(formula=formula)
                break
            except Exception:
                color_print(f'[Error] Invalid formula: {formula}, please double check and try again.\n', 'red')

    except (KeyboardInterrupt, EOFError):
        color_print('\n[Error] Input cancelled. Returning to previous menu...\n', 'red')
        time.sleep(1)
        return

    result = tools.generate_vasp_poscar(formula=formula)
    color_print(result['message'], 'green')
    time.sleep(3)

@register('1.1.2', 'Convert POSCAR coordinates (Direct <-> Cartesian).')
def command_1_1_2():
    try:
        while True:
            clear_and_print_entry_message()
            choices = [
                'Direct coordinates     —>  Cartesian coordinates',
                'Cartesian coordinates  —>  Direct coordinates',
            ] + global_commands()
            cli = Bullet(prompt='\n', choices=choices, margin=1, bullet=' ●', word_color=colors.foreground['green'])
            user_input = cli.launch()

            if user_input.startswith('AI'):
                from masgent.ai_mode import ai_backend
                ai_backend.main()
            elif user_input.startswith('New'):
                start_new_session()
            elif user_input.startswith('Back'):
                return
            elif user_input.startswith('Main'):
                run_command('0')
            elif user_input.startswith('Help'):
                print_help()
            elif user_input.startswith('Exit'):
                exit_and_cleanup()
            elif user_input.startswith('Direct coordinates'):
                to_cartesian = True
                break
            elif user_input.startswith('Cartesian coordinates'):
                to_cartesian = False
                break
            else:
                continue
    
    except (KeyboardInterrupt, EOFError):
        exit_and_cleanup()

    try:
        poscar_path = check_poscar()
    except (KeyboardInterrupt, EOFError):
        color_print('\n[Error] Input cancelled. Returning to previous menu...\n', 'red')
        time.sleep(1)
        return

    result = tools.convert_poscar_coordinates(poscar_path=poscar_path, to_cartesian=to_cartesian)
    color_print(result['message'], 'green')
    time.sleep(3)

@register('1.1.3', 'Convert structure file formats (CIF, POSCAR, XYZ).')
def command_1_1_3():
    try:
        while True:
            clear_and_print_entry_message()
            choices = [
                'POSCAR  ->  CIF',
                'POSCAR  ->  XYZ',
                'CIF     ->  POSCAR',
                'CIF     ->  XYZ',
                'XYZ     ->  POSCAR',
                'XYZ     ->  CIF',
            ] + global_commands()
            cli = Bullet(prompt='\n', choices=choices, margin=1, bullet=' ●', word_color=colors.foreground['green'])
            user_input = cli.launch()

            if user_input.startswith('AI'):
                from masgent.ai_mode import ai_backend
                ai_backend.main()
            elif user_input.startswith('New'):
                start_new_session()
            elif user_input.startswith('Back'):
                return
            elif user_input.startswith('Main'):
                run_command('0')
            elif user_input.startswith('Help'):
                print_help()
            elif user_input.startswith('Exit'):
                exit_and_cleanup()
            elif user_input.startswith('POSCAR') and user_input.endswith('CIF'):
                input_format, output_format = 'POSCAR', 'CIF'
                break
            elif user_input.startswith('POSCAR') and user_input.endswith('XYZ'):
                input_format, output_format = 'POSCAR', 'XYZ'
                break
            elif user_input.startswith('CIF') and user_input.endswith('POSCAR'):
                input_format, output_format = 'CIF', 'POSCAR'
                break
            elif user_input.startswith('CIF') and user_input.endswith('XYZ'):
                input_format, output_format = 'CIF', 'XYZ'
                break
            elif user_input.startswith('XYZ') and user_input.endswith('POSCAR'):
                input_format, output_format = 'XYZ', 'POSCAR'
                break
            elif user_input.startswith('XYZ') and user_input.endswith('CIF'):
                input_format, output_format = 'XYZ', 'CIF'
                break
            else:
                continue
    
    except (KeyboardInterrupt, EOFError):
        exit_and_cleanup()

    try:
        while True:
            input_path = color_input('\nEnter path to input structure file: ', 'yellow').strip()

            if not input_path:
                continue
            
            try:
                schemas.ConvertStructureFormatSchema(input_path=input_path, input_format=input_format, output_format=output_format)
                break
            except Exception:
                color_print(f'[Error] Invalid input: {input_path}, please double check and try again.\n', 'red')

    except (KeyboardInterrupt, EOFError):
        color_print('\n[Error] Input cancelled. Returning to previous menu...\n', 'red')
        time.sleep(1)
        return

    result = tools.convert_structure_format(input_path=input_path, input_format=input_format, output_format=output_format)
    color_print(result['message'], 'green')
    time.sleep(3)

@register('1.1.4', 'Generate structure with defects (Vacancy, Interstitial, Substitution).')
def command_1_1_4():
    try:
        while True:
            clear_and_print_entry_message()
            choices = [
                'Vacancy                 ->  Randomly remove atoms of a selected element',
                'Substitution            ->  Randomly substitute atoms of a selected element with defect element',
                'Interstitial (Voronoi)  ->  Add atom at interstitial sites using Voronoi method',
            ] + global_commands()
            cli = Bullet(prompt='\n', choices=choices, margin=1, bullet=' ●', word_color=colors.foreground['green'])
            user_input = cli.launch()

            if user_input.startswith('AI'):
                from masgent.ai_mode import ai_backend
                ai_backend.main()
            elif user_input.startswith('New'):
                start_new_session()
            elif user_input.startswith('Back'):
                return
            elif user_input.startswith('Main'):
                run_command('0')
            elif user_input.startswith('Help'):
                print_help()
            elif user_input.startswith('Exit'):
                exit_and_cleanup()
            elif user_input.startswith('Vacancy'):
                run_command('vacancy')
                break
            elif user_input.startswith('Interstitial (Voronoi)'):
                run_command('interstitial')
                break
            elif user_input.startswith('Substitution'):
                run_command('substitution')
                break
            else:
                continue
    
    except (KeyboardInterrupt, EOFError):
        exit_and_cleanup()

@register('vacancy', 'Generate structure with vacancy defects.')
def command_vacancy():
    try:
        poscar_path = check_poscar()
    except (KeyboardInterrupt, EOFError):
        color_print('\n[Error] Input cancelled. Returning to previous menu...\n', 'red')
        time.sleep(1)
        return

    try:
        while True:
            original_element = color_input('\nEnter the element to remove (e.g., Na): ', 'yellow').strip()
            if not original_element:
                continue
            
            try:
                schemas.CheckElement(element_symbol=original_element)
                schemas.CheckElementExistence(poscar_path=poscar_path, element_symbol=original_element)
                break
            except Exception:
                color_print(f'[Error] Invalid element {original_element}, please double check and try again.\n', 'red')
                continue

    except (KeyboardInterrupt, EOFError):
        color_print('\n[Error] Input cancelled. Returning to previous menu...\n', 'red')
        time.sleep(1)
        return

    try:
        while True:
            defect_amount_str = color_input('\nEnter the defect amount (fraction between 0 and 1, or atom count >=1): ', 'yellow').strip()
            if not defect_amount_str:
                continue

            try:
                if '.' in defect_amount_str:
                    defect_amount = float(defect_amount_str)
                else:
                    defect_amount = int(defect_amount_str)
                
                schemas.GenerateVaspPoscarWithVacancyDefects(poscar_path=poscar_path, original_element=original_element, defect_amount=defect_amount)
                break

            except Exception:
                color_print(f'[Error] Invalid defect amount: {defect_amount_str}, please double check and try again.\n', 'red')

    except (KeyboardInterrupt, EOFError):
        color_print('\n[Error] Input cancelled. Returning to previous menu...\n', 'red')
        time.sleep(1)
        return

    result = tools.generate_vasp_poscar_with_vacancy_defects(poscar_path=poscar_path, original_element=original_element, defect_amount=defect_amount)
    color_print(result['message'], 'green')
    time.sleep(3)

@register('substitution', 'Generate structure with substitution defects.')
def command_substitution():
    try:
        poscar_path = check_poscar()
    except (KeyboardInterrupt, EOFError):
        color_print('\n[Error] Input cancelled. Returning to previous menu...\n', 'red')
        time.sleep(1)
        return
    
    try:
        while True:
            original_element = color_input('\nEnter the target element to be substituted (e.g., Na): ', 'yellow').strip()
            if not original_element:
                continue
            
            try:
                schemas.CheckElement(element_symbol=original_element)
                schemas.CheckElementExistence(poscar_path=poscar_path, element_symbol=original_element)
                break
            
            except Exception:
                color_print(f'[Error] Invalid element {original_element}, please double check and try again.\n', 'red')
                
    except (KeyboardInterrupt, EOFError):
        color_print('\n[Error] Input cancelled. Returning to previous menu...\n', 'red')
        time.sleep(1)
        return
    
    try:
        while True:
            defect_element = color_input('\nEnter the defect element to substitute in (e.g., K): ', 'yellow').strip()
            if not defect_element:
                continue

            try:
                schemas.CheckElement(element_symbol=defect_element)
                break
            except Exception:
                color_print(f'[Error] Invalid element {defect_element}, please double check and try again.\n', 'red')

    except (KeyboardInterrupt, EOFError):
        color_print('\n[Error] Input cancelled. Returning to previous menu...\n', 'red')
        time.sleep(1)
        return

    try:
        while True:
            defect_amount_str = color_input('\nEnter the defect amount (fraction between 0 and 1, or atom count >=1): ', 'yellow').strip()
            if not defect_amount_str:
                continue
            
            try:
                if '.' in defect_amount_str:
                    defect_amount = float(defect_amount_str)
                else:
                    defect_amount = int(defect_amount_str)
                schemas.GenerateVaspPoscarWithSubstitutionDefects(poscar_path=poscar_path, original_element=original_element, defect_element=defect_element, defect_amount=defect_amount)
                break

            except Exception:
                color_print(f'[Error] Invalid defect amount: {defect_amount_str}, please double check and try again.\n', 'red')

    except (KeyboardInterrupt, EOFError):
        color_print('\n[Error] Input cancelled. Returning to previous menu...\n', 'red')
        time.sleep(1)
        return

    result = tools.generate_vasp_poscar_with_substitution_defects(poscar_path=poscar_path, original_element=original_element, defect_element=defect_element, defect_amount=defect_amount)
    color_print(result['message'], 'green')
    time.sleep(3)

@register('interstitial', 'Generate structure with interstitial (Voronoi) defects.')
def command_interstitial():
    try:
        poscar_path = check_poscar()
    except (KeyboardInterrupt, EOFError):
        color_print('\n[Error] Input cancelled. Returning to previous menu...\n', 'red')
        time.sleep(1)
        return
    
    try:
        while True:
            defect_element = color_input('\nEnter the defect element to add (e.g., Na): ', 'yellow').strip()
            if not defect_element:
                continue

            try:
                schemas.GenerateVaspPoscarWithInterstitialDefects(poscar_path=poscar_path, defect_element=defect_element)
                break
            except Exception:
                color_print(f'[Error] Invalid element {defect_element}, please double check and try again.\n', 'red')

    except (KeyboardInterrupt, EOFError):
        color_print('\n[Error] Input cancelled. Returning to previous menu...\n', 'red')
        time.sleep(1)
        return

    print('')
    with yaspin(Spinners.dots, text='Generating interstitial defects... See details in the log file.', color='cyan') as sp:
        result = tools.generate_vasp_poscar_with_interstitial_defects(poscar_path=poscar_path, defect_element=defect_element)
    color_print(result['message'], 'green')
    time.sleep(3)

@register('1.1.5', 'Generate supercell from POSCAR with specified scaling matrix.')
def command_1_1_5():
    try:
        poscar_path = check_poscar()
    except (KeyboardInterrupt, EOFError):
        color_print('\n[Error] Input cancelled. Returning to previous menu...\n', 'red')
        time.sleep(1)
        return
    
    try:
        while True:
            scaling_matrix = color_input('\nEnter the scaling matrix for 2x2x2 supercell (e.g., 2 0 0; 0 2 0; 0 0 2): ', 'yellow').strip()

            if not scaling_matrix:
                continue
            
            try:
                schemas.GenerateSupercellFromPoscar(poscar_path=poscar_path, scaling_matrix=scaling_matrix)
                break

            except Exception:
                color_print(f'[Error] Invalid scaling matrix: {scaling_matrix}, please double check and try again.\n', 'red')

    except (KeyboardInterrupt, EOFError):
        color_print('\n[Error] Input cancelled. Returning to previous menu...\n', 'red')
        time.sleep(1)
        return
    
    result = tools.generate_supercell_from_poscar(poscar_path=poscar_path, scaling_matrix=scaling_matrix)
    color_print(result['message'], 'green')
    time.sleep(3)

@register('1.1.6', 'Generate special quasi-random structure (SQS) from POSCAR.')
def command_1_1_6():
    try:
        poscar_path = check_poscar()
    except (KeyboardInterrupt, EOFError):
        color_print('\n[Error] Input cancelled. Returning to previous menu...\n', 'red')
        time.sleep(1)
        return
    
    try:
        while True:
            target_configurations_str = color_input('\nEnter target configurations (e.g., La: La=0.5,Y=0.5; Co: Al=0.75,Co=0.25): ', 'yellow').strip()
            
            if not target_configurations_str:
                continue

            try:
                # parse target configurations to {'La': {'La': 0.5, 'Y': 0.5}, 'Co': {'Al': 0.75, 'Co': 0.25}}
                target_configurations = {}
                for sublattice_str in target_configurations_str.split(';'):
                    sublattice_str = sublattice_str.strip()
                    if not sublattice_str:
                        continue
                    element, conc_str = sublattice_str.split(':')
                    element = element.strip()
                    conc_pairs = conc_str.split(',')
                    conc_dict = {}
                    for pair in conc_pairs:
                        species, conc = pair.split('=')
                        conc_dict[species.strip()] = float(conc.strip())
                    target_configurations[element] = conc_dict
                schemas.GenerateSqsFromPoscar(poscar_path=poscar_path, target_configurations=target_configurations)
                break
            except Exception as e:
                print(e)
                color_print(f'[Error] Invalid target configurations: {target_configurations_str}, please double check and try again.\n', 'red')
    
    except (KeyboardInterrupt, EOFError):
        color_print('\n[Error] Input cancelled. Returning to previous menu...\n', 'red')
        time.sleep(1)
        return
    
    try:
        while True:
            cutoffs_str = color_input('\nEnter cluster cutoffs in Å for pairs, triplets, and quadruplets (e.g., 8.0 4.0 4.0): ', 'yellow').strip()

            if not cutoffs_str:
                continue

            try:
                cutoffs = [float(x) for x in cutoffs_str.split()]
                schemas.GenerateSqsFromPoscar(poscar_path=poscar_path, target_configurations=target_configurations, cutoffs=cutoffs)
                break
            except Exception:
                color_print(f'[Error] Invalid cutoffs: {cutoffs_str}, please double check and try again.\n', 'red')

    except (KeyboardInterrupt, EOFError):
        color_print('\n[Error] Input cancelled. Returning to previous menu...\n', 'red')
        time.sleep(1)
        return
    
    try:
        while True:
            max_supercell_size_str = color_input('\nEnter maximum supercell size (e.g., 8): ', 'yellow').strip()

            if not max_supercell_size_str:
                continue

            try:
                max_supercell_size = int(max_supercell_size_str)
                schemas.GenerateSqsFromPoscar(poscar_path=poscar_path, target_configurations=target_configurations, cutoffs=cutoffs, max_supercell_size=max_supercell_size)
                break
            except Exception:
                color_print(f'[Error] Invalid maximum supercell size: {max_supercell_size_str}, please double check and try again.\n', 'red')

    except (KeyboardInterrupt, EOFError):
        color_print('\n[Error] Input cancelled. Returning to previous menu...\n', 'red')
        time.sleep(1)
        return
    
    try:
        while True:
            mc_steps_str = color_input('\nEnter number of Monte Carlo steps (e.g., >=1000): ', 'yellow').strip()

            if not mc_steps_str:
                continue

            try:
                mc_steps = int(mc_steps_str)
                schemas.GenerateSqsFromPoscar(poscar_path=poscar_path, target_configurations=target_configurations, cutoffs=cutoffs, max_supercell_size=max_supercell_size, mc_steps=mc_steps)
                break
            except Exception:
                color_print(f'[Error] Invalid number of Monte Carlo steps: {mc_steps_str}, please double check and try again.\n', 'red')

    except (KeyboardInterrupt, EOFError):
        color_print('\n[Error] Input cancelled. Returning to previous menu...\n', 'red')
        time.sleep(1)
        return
    
    print('')
    with yaspin(Spinners.dots, text='Generating SQS... See details in the log file.', color='cyan') as sp:
        result = tools.generate_sqs_from_poscar(
            poscar_path=poscar_path, 
            target_configurations=target_configurations, 
            cutoffs=cutoffs, 
            max_supercell_size=max_supercell_size, 
            mc_steps=mc_steps
        )
    color_print(result['message'], 'green')
    time.sleep(3)

@register('1.1.7', 'Generate surface slab from POSCAR with specified Miller indices, vacuum thickness, and slab layers.')
def command_1_1_7():
    try:
        poscar_path = check_poscar()
    except (KeyboardInterrupt, EOFError):
        color_print('\n[Error] Input cancelled. Returning to previous menu...\n', 'red')
        time.sleep(1)
        return
    
    try:
        while True:
            miller_indices_str = color_input('\nEnter the Miller indices (e.g., 1 1 1): ', 'yellow').strip()
            
            if not miller_indices_str:
                continue

            try:
                miller_indices = [int(x) for x in miller_indices_str.split()]
                schemas.GenerateSurfaceSlabFromPoscar(poscar_path=poscar_path, miller_indices=miller_indices)
                break
            except Exception:
                color_print(f'[Error] Invalid Miller indices: {miller_indices_str}, please double check and try again.\n', 'red')

    except (KeyboardInterrupt, EOFError):
        color_print('\n[Error] Input cancelled. Returning to previous menu...\n', 'red')
        time.sleep(1)
        return
    
    try:
        while True:
            vacuum_thickness_str = color_input('\nEnter the vacuum thickness in Å (e.g., 15.0): ', 'yellow').strip()

            if not vacuum_thickness_str:
                continue

            try:
                vacuum_thickness = float(vacuum_thickness_str)
                schemas.GenerateSurfaceSlabFromPoscar(poscar_path=poscar_path, miller_indices=miller_indices, vacuum_thickness=vacuum_thickness)
                break
            except Exception:
                color_print(f'[Error] Invalid vacuum thickness: {vacuum_thickness_str}, please double check and try again.\n', 'red')

    except (KeyboardInterrupt, EOFError):
        color_print('\n[Error] Input cancelled. Returning to previous menu...\n', 'red')
        time.sleep(1)
        return
    
    try:
        while True:
            slab_layers_str = color_input('\nEnter the number of slab layers (e.g., 4): ', 'yellow').strip()

            if not slab_layers_str:
                continue

            try:
                slab_layers = int(slab_layers_str)
                schemas.GenerateSurfaceSlabFromPoscar(poscar_path=poscar_path, miller_indices=miller_indices, vacuum_thickness=vacuum_thickness, slab_layers=slab_layers)
                break
            except Exception:
                color_print(f'[Error] Invalid slab layers: {slab_layers_str}, please double check and try again.\n', 'red')

    except (KeyboardInterrupt, EOFError):
        color_print('\n[Error] Input cancelled. Returning to previous menu...\n', 'red')
        time.sleep(1)
        return
    
    result = tools.generate_surface_slab_from_poscar(poscar_path=poscar_path, miller_indices=miller_indices, vacuum_thickness=vacuum_thickness, slab_layers=slab_layers)
    color_print(result['message'], 'green')
    time.sleep(3)

@register('1.1.8', 'Generate interface structure from two POSCAR files with specified parameters.')
def command_1_1_8():
    try:
        while True:
            lower_poscar_path = color_input('\nSelect the lower POSCAR file: ', 'yellow').strip()

            if not lower_poscar_path:
                continue

            try:
                schemas.CheckPoscar(poscar_path=lower_poscar_path)
                break
            except Exception:
                color_print(f'[Error] Invalid POSCAR: {lower_poscar_path}, please double check and try again.\n', 'red')

    except (KeyboardInterrupt, EOFError):
        color_print('\n[Error] Input cancelled. Returning to previous menu...\n', 'red')
        time.sleep(1)
        return
    
    try:
        while True:
            upper_poscar_path = color_input('\nSelect the upper POSCAR file: ', 'yellow').strip()

            if not upper_poscar_path:
                continue

            try:
                schemas.CheckPoscar(poscar_path=upper_poscar_path)
                break
            except Exception:
                color_print(f'[Error] Invalid POSCAR: {upper_poscar_path}, please double check and try again.\n', 'red')

    except (KeyboardInterrupt, EOFError):
        color_print('\n[Error] Input cancelled. Returning to previous menu...\n', 'red')
        time.sleep(1)
        return
    
    try:
        while True:
            hkl_str = color_input('\nEnter the Miller indices for the lower and upper surfaces (e.g., 1 0 0; 1 0 0): ', 'yellow').strip()

            if not hkl_str:
                continue

            try:
                lower_hkl = [int(x) for x in hkl_str.split(';')[0].strip().split()]
                upper_hkl = [int(x) for x in hkl_str.split(';')[1].strip().split()]
                schemas.GenerateInterfaceFromPoscars(lower_poscar_path=lower_poscar_path, upper_poscar_path=upper_poscar_path, lower_hkl=lower_hkl, upper_hkl=upper_hkl)
                break
            except Exception:
                color_print(f'[Error] Invalid Miller indices: {hkl_str}, please double check and try again.\n', 'red')
    
    except (KeyboardInterrupt, EOFError):
        color_print('\n[Error] Input cancelled. Returning to previous menu...\n', 'red')
        time.sleep(1)
        return
    
    try:
        while True:
            slab_layers_str = color_input('\nEnter the number of slab layers for the lower and upper slabs (e.g., 4 4): ', 'yellow').strip()

            if not slab_layers_str:
                continue

            try:
                lower_slab_layers = int(slab_layers_str.split()[0].strip())
                upper_slab_layers = int(slab_layers_str.split()[1].strip())
                schemas.GenerateInterfaceFromPoscars(lower_poscar_path=lower_poscar_path, upper_poscar_path=upper_poscar_path, lower_hkl=lower_hkl, upper_hkl=upper_hkl, lower_slab_layers=lower_slab_layers, upper_slab_layers=upper_slab_layers)
                break
            except Exception:
                color_print(f'[Error] Invalid slab layers: {slab_layers_str}, please double check and try again.\n', 'red')
    
    except (KeyboardInterrupt, EOFError):
        color_print('\n[Error] Input cancelled. Returning to previous menu...\n', 'red')
        time.sleep(1)
        return
    
    try:
        while True:
            slab_vacuum_str = color_input('\nEnter the vacuum thickness in Å (e.g., 15.0): ', 'yellow').strip()

            if not slab_vacuum_str:
                continue

            try:
                slab_vacuum = float(slab_vacuum_str)
                schemas.GenerateInterfaceFromPoscars(lower_poscar_path=lower_poscar_path, upper_poscar_path=upper_poscar_path, lower_hkl=lower_hkl, upper_hkl=upper_hkl, lower_slab_layers=lower_slab_layers, upper_slab_layers=upper_slab_layers, slab_vacuum=slab_vacuum)
                break
            except Exception:
                color_print(f'[Error] Invalid vacuum thickness: {slab_vacuum_str}, please double check and try again.\n', 'red')

    except (KeyboardInterrupt, EOFError):
        color_print('\n[Error] Input cancelled. Returning to previous menu...\n', 'red')
        time.sleep(1)
        return
    
    try:
        while True:
            area_str = color_input('\nEnter the minimum and maximum interface area to search in Å² (e.g., 50.0 500.0): ', 'yellow').strip()

            if not area_str:
                continue

            try:
                min_area = float(area_str.split()[0].strip())
                max_area = float(area_str.split()[1].strip())
                schemas.GenerateInterfaceFromPoscars(lower_poscar_path=lower_poscar_path, upper_poscar_path=upper_poscar_path, lower_hkl=lower_hkl, upper_hkl=upper_hkl, lower_slab_layers=lower_slab_layers, upper_slab_layers=upper_slab_layers, slab_vacuum=slab_vacuum, min_area=min_area, max_area=max_area)
                break
            except Exception:
                color_print(f'[Error] Invalid interface area: {area_str}, please double check and try again.\n', 'red')

    except (KeyboardInterrupt, EOFError):
        color_print('\n[Error] Input cancelled. Returning to previous menu...\n', 'red')
        time.sleep(1)
        return
    
    try:
        while True:
            interface_gap_str = color_input('\nEnter the interface gap in Å (e.g., 2.0): ', 'yellow').strip()

            if not interface_gap_str:
                continue

            try:
                interface_gap = float(interface_gap_str)
                schemas.GenerateInterfaceFromPoscars(lower_poscar_path=lower_poscar_path, upper_poscar_path=upper_poscar_path, lower_hkl=lower_hkl, upper_hkl=upper_hkl, lower_slab_layers=lower_slab_layers, upper_slab_layers=upper_slab_layers, slab_vacuum=slab_vacuum, min_area=min_area, max_area=max_area, interface_gap=interface_gap)
                break
            except Exception:
                color_print(f'[Error] Invalid interface gap: {interface_gap_str}, please double check and try again.\n', 'red')

    except (KeyboardInterrupt, EOFError):
        color_print('\n[Error] Input cancelled. Returning to previous menu...\n', 'red')
        time.sleep(1)
        return
    
    try:
        while True:
            tolerance_str = color_input('\nEnter the lattice vector tolerance (%) and angle tolerance (degrees) (e.g., 5.0 5.0): ', 'yellow').strip()

            if not tolerance_str:
                continue

            try:
                uv_tolerance = float(tolerance_str.split()[0].strip())
                angle_tolerance = float(tolerance_str.split()[1].strip())
                schemas.GenerateInterfaceFromPoscars(lower_poscar_path=lower_poscar_path, upper_poscar_path=upper_poscar_path, lower_hkl=lower_hkl, upper_hkl=upper_hkl, lower_slab_layers=lower_slab_layers, upper_slab_layers=upper_slab_layers, slab_vacuum=slab_vacuum, min_area=min_area, max_area=max_area, interface_gap=interface_gap, uv_tolerance=uv_tolerance, angle_tolerance=angle_tolerance)
                break
            except Exception:
                color_print(f'[Error] Invalid lattice vector tolerance: {tolerance_str}, please double check and try again.\n', 'red')
    
    except (KeyboardInterrupt, EOFError):
        color_print('\n[Error] Input cancelled. Returning to previous menu...\n', 'red')
        time.sleep(1)
        return
    
    try:
        while True:
            shape_filter_str = color_input('\nDo you want to apply shape filtering to only keep the most square-like interface? [y/n]: ', 'yellow').strip().lower()

            if not shape_filter_str:
                continue

            if shape_filter_str == 'y':
                shape_filter = True
                break
            elif shape_filter_str == 'n':
                shape_filter = False
                break
            else:
                continue

    except (KeyboardInterrupt, EOFError):
        color_print('\n[Error] Input cancelled. Returning to previous menu...\n', 'red')
        time.sleep(1)
        return
    
    print('')
    with yaspin(Spinners.dots, text='Generating interface structure... See details in the log file.', color='cyan') as sp:
        result = tools.generate_interface_from_poscars(
            lower_poscar_path=lower_poscar_path, 
            upper_poscar_path=upper_poscar_path, 
            lower_hkl=lower_hkl, 
            upper_hkl=upper_hkl, 
            lower_slab_layers=lower_slab_layers, 
            upper_slab_layers=upper_slab_layers, 
            slab_vacuum=slab_vacuum, 
            min_area=min_area, 
            max_area=max_area, 
            interface_gap=interface_gap, 
            uv_tolerance=uv_tolerance, 
            angle_tolerance=angle_tolerance, 
            shape_filter=shape_filter
        )
    color_print(result['message'], 'green')
    time.sleep(3)

@register('1.2.1', 'Prepare full VASP input files (INCAR, KPOINTS, POTCAR, POSCAR).')
def command_1_2_1():
    try:
        while True:
            clear_and_print_entry_message()
            choices = [
                'MPMetalRelaxSet  ->   suggested for metallic structure relaxation',
                'MPRelaxSet       ->   suggested for structure relaxation',
                'MPStaticSet      ->   suggested for static calculations',
                'MPNonSCFBandSet  ->   suggested for non-self-consistent field calculations (Band structure)',
                'MPNonSCFDOSSet   ->   suggested for non-self-consistent field calculations (Density of States)',
                'MPMDSet          ->   suggested for molecular dynamics simulations',
            ] + global_commands()
            cli = Bullet(prompt='\n', choices=choices, margin=1, bullet=' ●', word_color=colors.foreground['green'])
            user_input = cli.launch()

            if user_input.startswith('AI'):
                from masgent.ai_mode import ai_backend
                ai_backend.main()
            elif user_input.startswith('New'):
                start_new_session()
            elif user_input.startswith('Back'):
                return
            elif user_input.startswith('Main'):
                run_command('0')
            elif user_input.startswith('Help'):
                print_help()
            elif user_input.startswith('Exit'):
                exit_and_cleanup()
            elif user_input.startswith('MPMetalRelaxSet'):
                vasp_input_sets = 'MPMetalRelaxSet'
                break
            elif user_input.startswith('MPRelaxSet'):
                vasp_input_sets = 'MPRelaxSet'
                break
            elif user_input.startswith('MPStaticSet'):
                vasp_input_sets = 'MPStaticSet'
                break
            elif user_input.startswith('MPNonSCFBandSet'):
                vasp_input_sets = 'MPNonSCFBandSet'
                break
            elif user_input.startswith('MPNonSCFDOSSet'):
                vasp_input_sets = 'MPNonSCFDOSSet'
                break
            elif user_input.startswith('MPMDSet'):
                vasp_input_sets = 'MPMDSet'
                break
            else:
                continue
    
    except (KeyboardInterrupt, EOFError):
        exit_and_cleanup()
    
    try:
        poscar_path = check_poscar()
    except (KeyboardInterrupt, EOFError):
        color_print('\n[Error] Input cancelled. Returning to previous menu...\n', 'red')
        time.sleep(1)
        return

    result = tools.generate_vasp_inputs_from_poscar(poscar_path=poscar_path, vasp_input_sets=vasp_input_sets, only_incar=False)
    color_print(result['message'], 'green')
    time.sleep(3)

@register('1.2.2', 'Generate INCAR templates (relaxation, static, etc.).')
def command_1_2_2():
    try:
        while True:
            clear_and_print_entry_message()
            choices = [
                'MPMetalRelaxSet  ->   suggested for metallic structure relaxation',
                'MPRelaxSet       ->   suggested for structure relaxation',
                'MPStaticSet      ->   suggested for static calculations',
                'MPNonSCFBandSet  ->   suggested for non-self-consistent field calculations (Band structure)',
                'MPNonSCFDOSSet   ->   suggested for non-self-consistent field calculations (Density of States)',
                'MPMDSet          ->   suggested for molecular dynamics simulations',
            ] + global_commands()
            cli = Bullet(prompt='\n', choices=choices, margin=1, bullet=' ●', word_color=colors.foreground['green'])
            user_input = cli.launch()

            if user_input.startswith('AI'):
                from masgent.ai_mode import ai_backend
                ai_backend.main()
            elif user_input.startswith('New'):
                start_new_session()
            elif user_input.startswith('Back'):
                return
            elif user_input.startswith('Main'):
                run_command('0')
            elif user_input.startswith('Help'):
                print_help()
            elif user_input.startswith('Exit'):
                exit_and_cleanup()
            elif user_input.startswith('MPMetalRelaxSet'):
                vasp_input_sets = 'MPMetalRelaxSet'
                break
            elif user_input.startswith('MPRelaxSet'):
                vasp_input_sets = 'MPRelaxSet'
                break
            elif user_input.startswith('MPStaticSet'):
                vasp_input_sets = 'MPStaticSet'
                break
            elif user_input.startswith('MPNonSCFBandSet'):
                vasp_input_sets = 'MPNonSCFBandSet'
                break
            elif user_input.startswith('MPNonSCFDOSSet'):
                vasp_input_sets = 'MPNonSCFDOSSet'
                break
            elif user_input.startswith('MPMDSet'):
                vasp_input_sets = 'MPMDSet'
                break
            elif user_input.startswith('NEBSet'):
                vasp_input_sets = 'NEBSet'
                break
            else:
                continue
    
    except (KeyboardInterrupt, EOFError):
        exit_and_cleanup()
    
    try:
        poscar_path = check_poscar()
    except (KeyboardInterrupt, EOFError):
        color_print('\n[Error] Input cancelled. Returning to previous menu...\n', 'red')
        time.sleep(1)
        return

    result = tools.generate_vasp_inputs_from_poscar(poscar_path=poscar_path, vasp_input_sets=vasp_input_sets, only_incar=True)
    color_print(result['message'], 'green')
    time.sleep(3)

@register('1.2.3', 'Generate KPOINTS with specified accuracy.')
def command_1_2_3():
    try:
        while True:
            clear_and_print_entry_message()
            choices = [
                'Low     ->  Suitable for preliminary calculations, grid density = 1000 / number of atoms',
                'Medium  ->  Balanced accuracy and computational cost, grid density = 3000 / number of atoms',
                'High    ->  High accuracy for production runs, grid density = 5000 / number of atoms',
            ] + global_commands()
            cli = Bullet(prompt='\n', choices=choices, margin=1, bullet=' ●', word_color=colors.foreground['green'])
            user_input = cli.launch()

            if user_input.startswith('AI'):
                from masgent.ai_mode import ai_backend
                ai_backend.main()
            elif user_input.startswith('New'):
                start_new_session()
            elif user_input.startswith('Back'):
                return
            elif user_input.startswith('Main'):
                run_command('0')
            elif user_input.startswith('Help'):
                print_help()
            elif user_input.startswith('Exit'):
                exit_and_cleanup()
            elif user_input.startswith('Low'):
                accuracy_level = 'Low'
                break
            elif user_input.startswith('Medium'):
                accuracy_level = 'Medium'
                break
            elif user_input.startswith('High'):
                accuracy_level = 'High'
                break
            else:
                continue
    
    except (KeyboardInterrupt, EOFError):
        exit_and_cleanup()

    try:
        poscar_path = check_poscar()
    except (KeyboardInterrupt, EOFError):
        color_print('\n[Error] Input cancelled. Returning to previous menu...\n', 'red')
        time.sleep(1)
        return

    result = tools.customize_vasp_kpoints_with_accuracy(poscar_path=poscar_path, accuracy_level=accuracy_level)
    color_print(result['message'], 'green')
    time.sleep(3)

@register('1.2.4', 'Generate HPC Slurm job submission script for VASP calculations.')
def command_1_2_4():
    try:
        partition = color_input('\nEnter the HPC partition/queue name (default: normal): ', 'yellow').strip() or 'normal'
    except (KeyboardInterrupt, EOFError):
        color_print('\n[Error] Input cancelled. Returning to previous menu...\n', 'red')
        time.sleep(1)
        return

    try:
        while True:
            nodes = color_input('\nEnter the number of nodes (default: 1): ', 'yellow').strip()
            try:
                if not nodes:
                    nodes = 1
                else:
                    nodes = int(nodes)
                schemas.GenerateVaspInputsHpcSlurmScript(nodes=nodes)
                break
            except ValueError:
                color_print(f"\n[Error] Invalid number of nodes: {nodes}. Please enter a positive integer.\n", 'red')
                continue
    except (KeyboardInterrupt, EOFError):
        color_print('\n[Error] Input cancelled. Returning to previous menu...\n', 'red')
        time.sleep(1)
        return
    
    try:
        while True:
            ntasks = color_input('\nEnter the number of tasks per node (default: 8): ', 'yellow').strip()
            try:
                if not ntasks:
                    ntasks = 8
                else:
                    ntasks = int(ntasks)
                schemas.GenerateVaspInputsHpcSlurmScript(ntasks=ntasks)
                break
            except ValueError:
                color_print(f"\n[Error] Invalid number of tasks: {ntasks}. Please enter a positive integer.\n", 'red')
                continue
    except (KeyboardInterrupt, EOFError):
        color_print('\n[Error] Input cancelled. Returning to previous menu...\n', 'red')
        time.sleep(1)
        return
    
    try:
        while True:
            walltime = color_input('\nEnter the job walltime in format HH:MM:SS (default: 01:00:00): ', 'yellow').strip()
            try:
                if not walltime:
                    walltime = '01:00:00'
                schemas.GenerateVaspInputsHpcSlurmScript(walltime=walltime)
                break
            except ValueError:
                color_print(f"\n[Error] Invalid walltime format: {walltime}. Please use 'HH:MM:SS' format.\n", 'red')
                continue
    except (KeyboardInterrupt, EOFError):
        color_print('\n[Error] Input cancelled. Returning to previous menu...\n', 'red')
        time.sleep(1)
        return

    try:
        jobname = color_input('\nEnter the job name (default: masgent_job): ', 'yellow').strip()
        if not jobname:
            jobname = 'masgent_job'
    except (KeyboardInterrupt, EOFError):
        color_print('\n[Error] Input cancelled. Returning to previous menu...\n', 'red')
        time.sleep(1)
        return
    
    result = tools.generate_vasp_inputs_hpc_slurm_script(
        partition=partition,
        nodes=nodes,
        ntasks=ntasks,
        walltime=walltime,
        jobname=jobname,
        command='srun vasp_std > vasp.out'
    )
    color_print(result['message'], 'green')
    time.sleep(3)

@register('1.3.1', 'Generate VASP workflow for convergence tests for k-points and energy cutoff based on given POSCAR.')
def command_1_3_1():
    try:
        while True:
            clear_and_print_entry_message()
            choices = [
                'All      ->  Convergence tests for both energy cutoff and k-points',
                'ENCUT    ->  Convergence test for energy cutoff only',
                'KPOINTS  ->  Convergence test for k-points only',
            ] + global_commands()

            cli = Bullet(prompt='\n', choices=choices, margin=1, bullet=' ●', word_color=colors.foreground['green'])
            user_input = cli.launch()

            if user_input.startswith('AI'):
                from masgent.ai_mode import ai_backend
                ai_backend.main()
            elif user_input.startswith('New'):
                start_new_session()
            elif user_input.startswith('Back'):
                return
            elif user_input.startswith('Main'):
                run_command('0')
            elif user_input.startswith('Help'):
                print_help()
            elif user_input.startswith('Exit'):
                exit_and_cleanup()
            elif user_input.startswith('All'):
                test_type = 'all'
                break
            elif user_input.startswith('ENCUT'):
                test_type = 'encut'
                break
            elif user_input.startswith('KPOINTS'):
                test_type = 'kpoints'
                break
            else:
                continue

    except (KeyboardInterrupt, EOFError):
        exit_and_cleanup()
    
    try:
        poscar_path = check_poscar()
    except (KeyboardInterrupt, EOFError):
        color_print('\n[Error] Input cancelled. Returning to previous menu...\n', 'red')
        time.sleep(1)
        return

    if test_type == 'encut':
        try:
            while True:
                encut_levels_str = color_input('\nEnter the energy cutoff levels you want to test (e.g., 300 400 500 600 700): ', 'yellow').strip()

                if not encut_levels_str:
                    continue

                try:
                    encut_levels = [int(x) for x in encut_levels_str.split()]
                    schemas.GenerateVaspWorkflowOfConvergenceTests(poscar_path=poscar_path, test_type=test_type, encut_levels=encut_levels)
                    break
                except Exception:
                    color_print(f'[Error] Invalid energy cutoff levels: {encut_levels_str}, please double check and try again.\n', 'red')
        
        except (KeyboardInterrupt, EOFError):
            color_print('\n[Error] Input cancelled. Returning to previous menu...\n', 'red')
            time.sleep(1)
            return
        
    elif test_type == 'kpoints':
        try:
            while True:
                kpoint_levels_str = color_input('\nEnter the k-point grid density levels you want to test (e.g., 1000 2000 3000 4000 5000): ', 'yellow').strip()

                if not kpoint_levels_str:
                    continue

                try:
                    kpoint_levels = [int(x) for x in kpoint_levels_str.split()]
                    schemas.GenerateVaspWorkflowOfConvergenceTests(poscar_path=poscar_path, test_type=test_type, kpoint_levels=kpoint_levels)
                    break
                except Exception:
                    color_print(f'[Error] Invalid k-point levels: {kpoint_levels_str}, please double check and try again.\n', 'red')
        
        except (KeyboardInterrupt, EOFError):
            color_print('\n[Error] Input cancelled. Returning to previous menu...\n', 'red')
            time.sleep(1)
            return
        
    else:
        try:
            while True:
                encut_levels_str = color_input('\nEnter the energy cutoff levels you want to test (e.g., 300 400 500 600 700): ', 'yellow').strip()

                if not encut_levels_str:
                    continue

                try:
                    encut_levels = [int(x) for x in encut_levels_str.split()]
                    schemas.GenerateVaspWorkflowOfConvergenceTests(poscar_path=poscar_path, test_type=test_type, encut_levels=encut_levels)
                    break
                except Exception:
                    color_print(f'[Error] Invalid energy cutoff levels: {encut_levels_str}, please double check and try again.\n', 'red')
        
        except (KeyboardInterrupt, EOFError):
            color_print('\n[Error] Input cancelled. Returning to previous menu...\n', 'red')
            time.sleep(1)
            return
        
        try:
            while True:
                kpoint_levels_str = color_input('\nEnter the k-point grid density levels you want to test (e.g., 1000 2000 3000 4000 5000): ', 'yellow').strip()

                if not kpoint_levels_str:
                    continue

                try:
                    kpoint_levels = [int(x) for x in kpoint_levels_str.split()]
                    schemas.GenerateVaspWorkflowOfConvergenceTests(poscar_path=poscar_path, test_type=test_type, kpoint_levels=kpoint_levels)
                    break
                except Exception:
                    color_print(f'[Error] Invalid k-point levels: {kpoint_levels_str}, please double check and try again.\n', 'red')
        
        except (KeyboardInterrupt, EOFError):
            color_print('\n[Error] Input cancelled. Returning to previous menu...\n', 'red')
            time.sleep(1)
            return
    
    print('')
    with yaspin(Spinners.dots, text='Generating VASP workflow for convergence tests...', color='cyan') as sp:
        if test_type == 'encut':
            result = tools.generate_vasp_workflow_of_convergence_tests(poscar_path=poscar_path, test_type=test_type, encut_levels=encut_levels)
        elif test_type == 'kpoints':
            result = tools.generate_vasp_workflow_of_convergence_tests(poscar_path=poscar_path, test_type=test_type, kpoint_levels=kpoint_levels)
        else:
            result = tools.generate_vasp_workflow_of_convergence_tests(poscar_path=poscar_path, test_type=test_type, encut_levels=encut_levels, kpoint_levels=kpoint_levels)
    color_print(result['message'], 'green')
    time.sleep(3)

@register('1.3.2', 'Generate VASP workflow of equation of state (EOS) calculations based on given POSCAR.')
def command_1_3_2():
    try:
        poscar_path = check_poscar()
    except (KeyboardInterrupt, EOFError):
        color_print('\n[Error] Input cancelled. Returning to previous menu...\n', 'red')
        time.sleep(1)
        return

    try:
        while True:
            scale_factors_str = color_input('\nEnter the volume scale factors for EOS calculations (e.g., 0.94 0.96 0.98 1.00 1.02 1.04 1.06): ', 'yellow').strip()

            if not scale_factors_str:
                continue

            try:
                scale_factors = [float(x) for x in scale_factors_str.split()]
                schemas.GenerateVaspWorkflowOfEos(poscar_path=poscar_path, scale_factors=scale_factors)
                break
            except Exception:
                color_print(f'[Error] Invalid scale factors: {scale_factors_str}, please double check and try again.\n', 'red')

    except (KeyboardInterrupt, EOFError):
        color_print('\n[Error] Input cancelled. Returning to previous menu...\n', 'red')
        time.sleep(1)
        return
    
    print('')
    with yaspin(Spinners.dots, text='Generating VASP workflow for EOS calculations...', color='cyan') as sp:
        result = tools.generate_vasp_workflow_of_eos(poscar_path=poscar_path, scale_factors=scale_factors)
    color_print(result['message'], 'green')
    time.sleep(3)

@register('1.3.3', 'Generate VASP workflow for elastic constants calculations based on given POSCAR.')
def command_1_3_3():
    try:
        poscar_path = check_poscar()
    except (KeyboardInterrupt, EOFError):
        color_print('\n[Error] Input cancelled. Returning to previous menu...\n', 'red')
        time.sleep(1)
        return

    print('')
    with yaspin(Spinners.dots, text='Generating VASP workflow for elastic constants calculations...', color='cyan') as sp:
        result = tools.generate_vasp_workflow_of_elastic_constants(poscar_path=poscar_path)
    color_print(result['message'], 'green')
    time.sleep(3)

@register('1.3.4', 'Generate VASP workflow for ab initio molecular dynamics (AIMD) simulations based on given POSCAR.')
def command_1_3_4():
    try:
        poscar_path = check_poscar()
    except (KeyboardInterrupt, EOFError):
        color_print('\n[Error] Input cancelled. Returning to previous menu...\n', 'red')
        time.sleep(1)
        return

    try:
        while True:
            temperatures_str = color_input('\nEnter the simulation temperature(s) in K (e.g., 500 1000 1500 2000 2500): ', 'yellow').strip()

            if not temperatures_str:
                continue

            try:
                temperatures = [int(x) for x in temperatures_str.split()]
                schemas.GenerateVaspWorkflowOfAimd(poscar_path=poscar_path, temperatures=temperatures)
                break
            except Exception:
                color_print(f'[Error] Invalid temperature: {temperatures_str}, please double check and try again.\n', 'red')

    except (KeyboardInterrupt, EOFError):
        color_print('\n[Error] Input cancelled. Returning to previous menu...\n', 'red')
        time.sleep(1)
        return
    
    try:
        while True:
            md_steps_str = color_input('\nEnter the number of MD steps (e.g., 1000): ', 'yellow').strip()

            if not md_steps_str:
                continue

            try:
                md_steps = int(md_steps_str)
                schemas.GenerateVaspWorkflowOfAimd(poscar_path=poscar_path, temperatures=temperatures, md_steps=md_steps)
                break
            except Exception:
                color_print(f'[Error] Invalid MD steps: {md_steps_str}, please double check and try again.\n', 'red')

    except (KeyboardInterrupt, EOFError):
        color_print('\n[Error] Input cancelled. Returning to previous menu...\n', 'red')
        time.sleep(1)
        return
    
    try:
        while True:
            md_timestep_str = color_input('\nEnter the MD timestep in fs (e.g., 2.0): ', 'yellow').strip()

            if not md_timestep_str:
                continue

            try:
                md_timestep = float(md_timestep_str)
                schemas.GenerateVaspWorkflowOfAimd(poscar_path=poscar_path, temperatures=temperatures, md_steps=md_steps, md_timestep=md_timestep)
                break
            except Exception:
                color_print(f'[Error] Invalid MD timestep: {md_timestep_str}, please double check and try again.\n', 'red')

    except (KeyboardInterrupt, EOFError):
        color_print('\n[Error] Input cancelled. Returning to previous menu...\n', 'red')
        time.sleep(1)
        return
    
    print('')
    with yaspin(Spinners.dots, text='Generating VASP workflow for AIMD simulations...', color='cyan') as sp:
        result = tools.generate_vasp_workflow_of_aimd(poscar_path=poscar_path, temperatures=temperatures, md_steps=md_steps, md_timestep=md_timestep)
    color_print(result['message'], 'green')
    time.sleep(3)

@register('1.3.5', 'Generate VASP workflow for Nudged Elastic Band (NEB) calculations based on given initial and final POSCARs.')
def command_1_3_5():
    try:
        while True:
            initial_poscar_path = color_input('\nEnter the path to the initial POSCAR file: ', 'yellow').strip()
            
            if not initial_poscar_path:
                continue

            try:
                schemas.CheckPoscar(poscar_path=initial_poscar_path)
                break
            except Exception:
                color_print(f'[Error] Invalid POSCAR: {initial_poscar_path}, please double check and try again.\n', 'red')
    
    except (KeyboardInterrupt, EOFError):
        color_print('\n[Error] Input cancelled. Returning to previous menu...\n', 'red')
        time.sleep(1)
        return
    
    try:
        while True:
            final_poscar_path = color_input('\nEnter the path to the final POSCAR file: ', 'yellow').strip()
            
            if not final_poscar_path:
                continue

            try:
                schemas.CheckPoscar(poscar_path=final_poscar_path)
                break
            except Exception:
                color_print(f'[Error] Invalid POSCAR: {final_poscar_path}, please double check and try again.\n', 'red')

    except (KeyboardInterrupt, EOFError):
        color_print('\n[Error] Input cancelled. Returning to previous menu...\n', 'red')
        time.sleep(1)
        return
    
    try:
        while True:
            num_images_str = color_input('\nEnter the number of intermediate images (e.g., 5): ', 'yellow').strip()

            if not num_images_str:
                continue

            try:
                num_images = int(num_images_str)
                schemas.GenerateVaspWorkflowOfNeb(initial_poscar_path=initial_poscar_path, final_poscar_path=final_poscar_path, num_images=num_images)
                break
            except Exception:
                color_print(f'[Error] Invalid number of images: {num_images_str}, please double check and try again.\n', 'red')
    
    except (KeyboardInterrupt, EOFError):
        color_print('\n[Error] Input cancelled. Returning to previous menu...\n', 'red')
        time.sleep(1)
        return
    
    print('')
    with yaspin(Spinners.dots, text='Generating VASP workflow for NEB calculations...', color='cyan') as sp:
        result = tools.generate_vasp_workflow_of_neb(initial_poscar_path=initial_poscar_path, final_poscar_path=final_poscar_path, num_images=num_images)
    color_print(result['message'], 'green')
    time.sleep(3)

@register('1.4.1', 'Convergence test analysis')
def command_1_4_1():
    try:
        while True:
            convergence_tests_dir = color_input('\nEnter the convergence tests directory path that contains encut and kpoints subdirectories: ', 'yellow').strip()
            
            if not convergence_tests_dir:
                continue

            try:
                os.path.exists(convergence_tests_dir)
                break
            except Exception:
                color_print(f'[Error] Invalid directory: {convergence_tests_dir}, please double check and try again.\n', 'red')

    except (KeyboardInterrupt, EOFError):
        color_print('\n[Error] Input cancelled. Returning to previous menu...\n', 'red')
        time.sleep(1)
        return
    
    print('')
    with yaspin(Spinners.dots, text='Analyzing VASP convergence tests...', color='cyan') as sp:
        result = tools.analyze_vasp_workflow_of_convergence_tests(convergence_tests_dir=convergence_tests_dir)
    color_print(result['message'], 'green')
    time.sleep(3)

@register('1.4.2', 'Equation of State (EOS) analysis')
def command_1_4_2():
    try:
        while True:
            eos_dir = color_input('\nEnter the EOS calculations directory path that contains volume-scaled subdirectories: ', 'yellow').strip()
            
            if not eos_dir:
                continue

            try:
                os.path.exists(eos_dir)
                break
            except Exception:
                color_print(f'[Error] Invalid directory: {eos_dir}, please double check and try again.\n', 'red')

    except (KeyboardInterrupt, EOFError):
        color_print('\n[Error] Input cancelled. Returning to previous menu...\n', 'red')
        time.sleep(1)
        return
    
    print('')
    with yaspin(Spinners.dots, text='Analyzing VASP EOS calculations...', color='cyan') as sp:
        result = tools.analyze_vasp_workflow_of_eos(eos_dir=eos_dir)
    color_print(result['message'], 'green')
    time.sleep(3)

@register('1.4.3', 'Elastic constants analysis')
def command_1_4_3():
    try:
        while True:
            elastic_constants_dir = color_input('\nEnter the elastic constants calculations directory path that contains strain subdirectories: ', 'yellow').strip()

            if not elastic_constants_dir:
                continue

            try:
                os.path.exists(elastic_constants_dir)
                break
            except Exception:
                color_print(f'[Error] Invalid directory: {elastic_constants_dir}, please double check and try again.\n', 'red')

    except (KeyboardInterrupt, EOFError):
        color_print('\n[Error] Input cancelled. Returning to previous menu...\n', 'red')
        time.sleep(1)
        return
    
    print('')
    with yaspin(Spinners.dots, text='Analyzing VASP elastic constants calculations...', color='cyan') as sp:
        result = tools.analyze_vasp_workflow_of_elastic_constants(elastic_constants_dir=elastic_constants_dir)
    color_print(result['message'], 'green')
    time.sleep(3)

@register('1.4.4', 'Ab initio molecular dynamics (AIMD) analysis')
def command_1_4_4():
    try:
        while True:
            aimd_dir = color_input('\nEnter the AIMD simulations directory path that contains MD temperature subdirectories: ', 'yellow').strip()

            if not aimd_dir:
                continue

            try:
                os.path.exists(aimd_dir)
                break
            except Exception:
                color_print(f'[Error] Invalid directory: {aimd_dir}, please double check and try again.\n', 'red')

    except (KeyboardInterrupt, EOFError):
        color_print('\n[Error] Input cancelled. Returning to previous menu...\n', 'red')
        time.sleep(1)
        return
    
    try:
        while True:
            specie = color_input('\nEnter the atomic specie symbol for MSD calculation (e.g., Li): ', 'yellow').strip()

            if not specie:
                continue

            try:
                schemas.CheckElement(element_symbol=specie)
                for root, dirs, files in os.walk(aimd_dir):
                    if 'POSCAR' in files:
                        poscar_path = os.path.join(root, 'POSCAR')
                        schemas.CheckElementExistence(poscar_path=poscar_path, element_symbol=specie)
                break
            except Exception:
                color_print(f'[Error] Invalid atomic specie symbol: {specie}, please double check and try again.\n', 'red')

    except (KeyboardInterrupt, EOFError):
        color_print('\n[Error] Input cancelled. Returning to previous menu...\n', 'red')
        time.sleep(1)
        return

    print('')
    with yaspin(Spinners.dots, text='Analyzing VASP AIMD simulations...', color='cyan') as sp:
        result = tools.analyze_vasp_workflow_of_aimd(aimd_dir=aimd_dir, specie=specie)
    color_print(result['message'], 'green')
    time.sleep(3)

def call_mlps(mlps_type: str):
    try:
        while True:
            clear_and_print_entry_message()
            choices = [
                '1. Single Point Energy Calculation',
                '2. Equation of State (EOS) Calculation',
                '3. Elastic Constants Calculation',
                '4. Molecular Dynamics Simulation (NVT)',
            ] + global_commands()
            cli = Bullet(prompt='\n', choices=choices, margin=1, bullet=' ●', word_color=colors.foreground['green'])
            user_input = cli.launch()
            if user_input.startswith('AI'):
                from masgent.ai_mode import ai_backend
                ai_backend.main()
            elif user_input.startswith('New'):
                start_new_session()
            elif user_input.startswith('Back'):
                return
            elif user_input.startswith('Main'):
                run_command('0')
            elif user_input.startswith('Help'):
                print_help()
            elif user_input.startswith('Exit'):
                exit_and_cleanup()
            elif user_input.startswith('1'):
                task_type = 'single'
                break
            elif user_input.startswith('2'):
                task_type = 'eos'
                break
            elif user_input.startswith('3'):
                task_type = 'elastic'
                break
            elif user_input.startswith('4'):
                task_type = 'md'
                break
            else:
                continue

    except (KeyboardInterrupt, EOFError):
        exit_and_cleanup()
    
    try:
        poscar_path = check_poscar()
    except (KeyboardInterrupt, EOFError):
        color_print('\n[Error] Input cancelled. Returning to previous menu...\n', 'red')
        time.sleep(1)
        return

    if task_type in ['single', 'eos', 'elastic']:
        try:
            while True:
                fmax_str = color_input('\nEnter the maximum force convergence criterion in eV/Å (default: 0.1): ', 'yellow').strip()

                if not fmax_str:
                    fmax = 0.1
                    break

                try:
                    fmax = float(fmax_str)
                    schemas.RunSimulationUsingMlps(poscar_path=poscar_path, fmax=fmax)
                    break
                except Exception:
                    color_print(f'[Error] Invalid force criterion: {fmax_str}, please double check and try again.\n', 'red')
        
        except (KeyboardInterrupt, EOFError):
            color_print('\n[Error] Input cancelled. Returning to previous menu...\n', 'red')
            time.sleep(1)
            return
        
        try:
            while True:
                max_steps_str = color_input('\nEnter the maximum number of optimization steps (default: 500): ', 'yellow').strip()

                if not max_steps_str:
                    max_steps = 500
                    break

                try:
                    max_steps = int(max_steps_str)
                    schemas.RunSimulationUsingMlps(poscar_path=poscar_path, max_steps=max_steps)
                    break
                except Exception:
                    color_print(f'[Error] Invalid maximum steps: {max_steps_str}, please double check and try again.\n', 'red')
        
        except (KeyboardInterrupt, EOFError):
            color_print('\n[Error] Input cancelled. Returning to previous menu...\n', 'red')
            time.sleep(1)
            return

        print('')
        with yaspin(Spinners.dots, text=f'Running simulation using {mlps_type}... See details in the log file. ', color='cyan') as sp:
            result = tools.run_simulation_using_mlps(poscar_path=poscar_path, mlps_type=mlps_type, task_type=task_type, fmax=fmax, max_steps=max_steps)
        color_print(result['message'], 'green')
        time.sleep(3)
    
    elif task_type == 'md':
        try:
            while True:
                temperature_str = color_input('\nEnter the simulation temperature in K (default: 1000): ', 'yellow').strip()

                if not temperature_str:
                    temperature = 1000
                    break

                try:
                    temperature = int(temperature_str)
                    schemas.RunSimulationUsingMlps(poscar_path=poscar_path, temperature=temperature)
                    break
                except Exception:
                    color_print(f'[Error] Invalid temperature: {temperature_str}, please double check and try again.\n', 'red')
        
        except (KeyboardInterrupt, EOFError):
            color_print('\n[Error] Input cancelled. Returning to previous menu...\n', 'red')
            time.sleep(1)
            return
        
        try:
            while True:
                md_steps_str = color_input('\nEnter the number of MD steps (default: 1000): ', 'yellow').strip()

                if not md_steps_str:
                    md_steps = 1000
                    break
                
                try:
                    md_steps = int(md_steps_str)
                    schemas.RunSimulationUsingMlps(poscar_path=poscar_path, md_steps=md_steps)
                    break
                except Exception:
                    color_print(f'[Error] Invalid MD steps: {md_steps_str}, please double check and try again.\n', 'red')

        except (KeyboardInterrupt, EOFError):
            color_print('\n[Error] Input cancelled. Returning to previous menu...\n', 'red')
            time.sleep(1)
            return
        
        try:
            while True:
                md_timestep_str = color_input('\nEnter the MD timestep in fs (default: 5.0 fs): ', 'yellow').strip()

                if not md_timestep_str:
                    md_timestep = 5.0
                    break

                try:
                    md_timestep = float(md_timestep_str)
                    schemas.RunSimulationUsingMlps(poscar_path=poscar_path, md_timestep=md_timestep)
                    break
                except Exception:
                    color_print(f'[Error] Invalid MD timestep: {md_timestep_str}, please double check and try again.\n', 'red')
        
        except (KeyboardInterrupt, EOFError):
            color_print('\n[Error] Input cancelled. Returning to previous menu...\n', 'red')
            time.sleep(1)
            return

        print('')
        with yaspin(Spinners.dots, text=f'Running simulation using {mlps_type}... See details in the log file. ', color='cyan') as sp:
            result = tools.run_simulation_using_mlps(poscar_path=poscar_path, mlps_type=mlps_type, task_type=task_type, temperature=temperature, md_steps=md_steps, md_timestep=md_timestep)
        color_print(result['message'], 'green')
        time.sleep(3)

@register('2.1', 'SevenNet')
def command_2_1():
    call_mlps(mlps_type='SevenNet')

@register('2.2', 'CHGNet')
def command_2_2():
    call_mlps(mlps_type='CHGNet')

@register('2.3', 'Orb-v3')
def command_2_3():
    call_mlps(mlps_type='Orb-v3')

@register('2.4', 'MatSim')
def command_2_4():
    call_mlps(mlps_type='MatSim')

@register('3.1.1', 'Feature analysis and visualization')
def command_3_1_1():
    try:
        while True:
            input_data_path = color_input('\nEnter the path to the input feature data file (CSV): ', 'yellow').strip()

            if not input_data_path:
                continue

            try:
                schemas.CheckCSVFile(file_path=input_data_path)
                break
            except Exception:
                color_print(f'[Error] Invalid CSV file: {input_data_path}, please double check and try again.\n', 'red')

    except (KeyboardInterrupt, EOFError):
        color_print('\n[Error] Input cancelled. Returning to previous menu...\n', 'red')
        time.sleep(1)
        return
    
    try:
        while True:
            output_data_path = color_input('\nEnter the path to the output feature data file (CSV): ', 'yellow').strip()

            if not output_data_path:
                continue

            try:
                schemas.CheckCSVFile(file_path=output_data_path)
                break
            except Exception:
                color_print(f'[Error] Invalid directory: {output_data_path}, please double check and try again.\n', 'red')

    except (KeyboardInterrupt, EOFError):
        color_print('\n[Error] Input cancelled. Returning to previous menu...\n', 'red')
        time.sleep(1)
        return
    
    result = tools.analyze_features_for_machine_learning(input_data_path=input_data_path, output_data_path=output_data_path)
    color_print(result['message'], 'green')
    time.sleep(3)

@register('3.1.2', 'Dimensionality reduction (if too many features)')
def command_3_1_2():
    try:
        while True:
            input_data_path = color_input('\nEnter the path to the input feature data file (CSV): ', 'yellow').strip()

            if not input_data_path:
                continue

            try:
                schemas.CheckCSVFile(file_path=input_data_path)
                break
            except Exception:
                color_print(f'[Error] Invalid CSV file: {input_data_path}, please double check and try again.\n', 'red')

    except (KeyboardInterrupt, EOFError):
        color_print('\n[Error] Input cancelled. Returning to previous menu...\n', 'red')
        time.sleep(1)
        return
    
    try:
        while True:
            n_components_str = color_input('\nEnter the number of principal components to reduce to (e.g., 2): ', 'yellow').strip()

            if not n_components_str:
                continue

            try:
                n_components = int(n_components_str)
                schemas.ReduceDimensionsForMachineLearning(input_data_path=input_data_path, n_components=n_components)
                break
            except Exception:
                color_print(f'[Error] Invalid number of components: {n_components_str}, please double check and try again.\n', 'red')

    except (KeyboardInterrupt, EOFError):
        color_print('\n[Error] Input cancelled. Returning to previous menu...\n', 'red')
        time.sleep(1)
        return
    
    result = tools.reduce_dimensions_for_machine_learning(input_data_path=input_data_path, n_components=n_components)
    color_print(result['message'], 'green')
    time.sleep(3)

@register('3.1.3', 'Data augmentation (if limited data)')
def command_3_1_3():
    try:
        while True:
            input_data_path = color_input('\nEnter the path to the input feature data file (CSV): ', 'yellow').strip()

            if not input_data_path:
                continue

            try:
                schemas.CheckCSVFile(file_path=input_data_path)
                break
            except Exception:
                color_print(f'[Error] Invalid CSV file: {input_data_path}, please double check and try again.\n', 'red')

    except (KeyboardInterrupt, EOFError):
        color_print('\n[Error] Input cancelled. Returning to previous menu...\n', 'red')
        time.sleep(1)
        return
    
    try:
        while True:
            output_data_path = color_input('\nEnter the path to the output feature data file (CSV): ', 'yellow').strip()

            if not output_data_path:
                continue

            try:
                schemas.CheckCSVFile(file_path=output_data_path)
                break
            except Exception:
                color_print(f'[Error] Invalid directory: {output_data_path}, please double check and try again.\n', 'red')

    except (KeyboardInterrupt, EOFError):   
        color_print('\n[Error] Input cancelled. Returning to previous menu...\n', 'red')
        time.sleep(1)
        return
    
    try:
        while True:
            num_augmentations_str = color_input('\nEnter the number of augmentations to perform (e.g., 100): ', 'yellow').strip()

            if not num_augmentations_str:
                continue

            try:
                num_augmentations = int(num_augmentations_str)
                schemas.AugmentDataForMachineLearning(input_data_path=input_data_path, output_data_path=output_data_path, num_augmentations=num_augmentations)
                break
            except Exception:
                color_print(f'[Error] Invalid number of augmentations: {num_augmentations_str}, please double check and try again.\n', 'red')
    
    except (KeyboardInterrupt, EOFError):
        color_print('\n[Error] Input cancelled. Returning to previous menu...\n', 'red')
        time.sleep(1)
        return
    
    result = tools.augment_data_for_machine_learning(input_data_path=input_data_path, output_data_path=output_data_path, num_augmentations=num_augmentations)
    color_print(result['message'], 'green')
    time.sleep(3)

@register('3.2', 'Model Design & Hyperparameter Tuning')
def command_3_2():
    try:
        while True:
            input_data_path = color_input('\nEnter the path to the input feature data file (CSV): ', 'yellow').strip()

            if not input_data_path:
                continue

            try:
                schemas.CheckCSVFile(file_path=input_data_path)
                break
            except Exception:
                color_print(f'[Error] Invalid CSV file: {input_data_path}, please double check and try again.\n', 'red')

    except (KeyboardInterrupt, EOFError):
        color_print('\n[Error] Input cancelled. Returning to previous menu...\n', 'red')
        time.sleep(1)
        return
    
    try:
        while True:
            output_data_path = color_input('\nEnter the path to the output feature data file (CSV): ', 'yellow').strip()

            if not output_data_path:
                continue

            try:
                schemas.CheckCSVFile(file_path=output_data_path)
                break
            except Exception:
                color_print(f'[Error] Invalid directory: {output_data_path}, please double check and try again.\n', 'red')

    except (KeyboardInterrupt, EOFError):
        color_print('\n[Error] Input cancelled. Returning to previous menu...\n', 'red')
        time.sleep(1)
        return
    
    try:
        while True:
            num_trials_str = color_input('\nEnter the number of hyperparameter tuning trials (e.g., 50): ', 'yellow').strip()

            if not num_trials_str:
                continue

            try:
                num_trials = int(num_trials_str)
                schemas.DesignModelForMachineLearning(input_data_path=input_data_path, output_data_path=output_data_path, n_trials=num_trials)
                break
            except Exception:
                color_print(f'[Error] Invalid number of trials: {num_trials_str}, please double check and try again.\n', 'red')

    except (KeyboardInterrupt, EOFError):
        color_print('\n[Error] Input cancelled. Returning to previous menu...\n', 'red')
        time.sleep(1)
        return
    
    result = tools.design_model_for_machine_learning(input_data_path=input_data_path, output_data_path=output_data_path, n_trials=num_trials)
    color_print(result['message'], 'green')
    time.sleep(3)

@register('3.3', 'Model Training & Evaluation')
def command_3_3():
    try:
        while True:
            input_data_path = color_input('\nEnter the path to the input feature data file (CSV): ', 'yellow').strip()

            if not input_data_path:
                continue

            try:
                schemas.CheckCSVFile(file_path=input_data_path)
                break
            except Exception:
                color_print(f'[Error] Invalid CSV file: {input_data_path}, please double check and try again.\n', 'red')

    except (KeyboardInterrupt, EOFError):
        color_print('\n[Error] Input cancelled. Returning to previous menu...\n', 'red')
        time.sleep(1)
        return
    
    try:
        while True:
            output_data_path = color_input('\nEnter the path to the output feature data file (CSV): ', 'yellow').strip()

            if not output_data_path:
                continue

            try:
                schemas.CheckCSVFile(file_path=output_data_path)
                break
            except Exception:
                color_print(f'[Error] Invalid directory: {output_data_path}, please double check and try again.\n', 'red')

    except (KeyboardInterrupt, EOFError):
        color_print('\n[Error] Input cancelled. Returning to previous menu...\n', 'red')
        time.sleep(1)
        return
    
    try:
        while True:
            best_model_path = color_input('\nEnter the path to the best model file from model design (e.g., best_model.pkl): ', 'yellow').strip()

            if not best_model_path:
                continue

            try:
                schemas.CheckPklFile(file_path=best_model_path)
                break
            except Exception:
                color_print(f'[Error] Invalid model file: {best_model_path}, please double check and try again.\n', 'red')

    except (KeyboardInterrupt, EOFError):
        color_print('\n[Error] Input cancelled. Returning to previous menu...\n', 'red')
        time.sleep(1)
        return
    
    try:
        while True:
            best_model_params_path = color_input('\nEnter the path to the best model hyperparameters file from model design (e.g., best_model_params.log): ', 'yellow').strip()

            if not best_model_params_path:
                continue

            try:
                schemas.CheckLogFile(file_path=best_model_params_path)
                break
            except Exception:
                color_print(f'[Error] Invalid hyperparameters file: {best_model_params_path}, please double check and try again.\n', 'red')

    except (KeyboardInterrupt, EOFError):
        color_print('\n[Error] Input cancelled. Returning to previous menu...\n', 'red')
        time.sleep(1)
        return
    
    try:
        while True:
            max_epochs_str = color_input('\nEnter the maximum number of training epochs (e.g., 1000): ', 'yellow').strip()

            if not max_epochs_str:
                continue

            try:
                max_epochs = int(max_epochs_str)
                schemas.TrainModelForMachineLearning(
                    input_data_path=input_data_path, 
                    output_data_path=output_data_path, 
                    best_model_path=best_model_path, 
                    best_model_params_path=best_model_params_path, 
                    max_epochs=max_epochs
                    )
                break
            except Exception:
                color_print(f'[Error] Invalid maximum epochs: {max_epochs_str}, please double check and try again.\n', 'red')

    except (KeyboardInterrupt, EOFError):
        color_print('\n[Error] Input cancelled. Returning to previous menu...\n', 'red')
        time.sleep(1)
        return
    
    try:
        while True:
            patience_str = color_input('\nEnter the early stopping patience (number of epochs with no improvement) (e.g., 50): ', 'yellow').strip()

            if not patience_str:
                continue

            try:
                patience = int(patience_str)
                schemas.TrainModelForMachineLearning(
                    input_data_path=input_data_path, 
                    output_data_path=output_data_path, 
                    best_model_path=best_model_path, 
                    best_model_params_path=best_model_params_path, 
                    max_epochs=max_epochs, 
                    patience=patience
                    )
                break
            except Exception:
                color_print(f'[Error] Invalid patience: {patience_str}, please double check and try again.\n', 'red')

    except (KeyboardInterrupt, EOFError):
        color_print('\n[Error] Input cancelled. Returning to previous menu...\n', 'red')
        time.sleep(1)
        return
    
    result = tools.train_model_for_machine_learning(
        input_data_path=input_data_path, 
        output_data_path=output_data_path, 
        best_model_path=best_model_path, 
        best_model_params_path=best_model_params_path, 
        max_epochs=max_epochs, 
        patience=patience
        )
    color_print(result['message'], 'green')
    time.sleep(3)



#!/usr/bin/env python
# -*-coding:utf-8 -*-

import os
import numpy as np
import pandas as pd
import matplotlib
matplotlib.use('Agg')
import matplotlib.pyplot as plt
import seaborn as sns
import torch
from sklearn.preprocessing import StandardScaler
from sklearn.model_selection import train_test_split
from sklearn.metrics import root_mean_squared_error, r2_score
from matplotlib.ticker import FormatStrFormatter
from tqdm import tqdm

def init_weights(m):
    if isinstance(m, torch.nn.Linear):
        m.reset_parameters()

def get_std(data):
    scaler = StandardScaler()
    data_std = scaler.fit_transform(data)
    return scaler, data_std

def train(input_data, output_data, best_model_pkl, best_model_params, epochs, patience, save_path):
    global INPUT_DATA, OUTPUT_DATA, BEST_MODEL_PKL, BEST_MODEL_PARAMS, EPOCHS, PATIENCE, SAVE_PATH, DEVICE
    
    INPUT_DATA = input_data
    OUTPUT_DATA = output_data
    BEST_MODEL_PKL = best_model_pkl
    BEST_MODEL_PARAMS = best_model_params
    EPOCHS = epochs
    PATIENCE = patience
    SAVE_PATH = save_path
    DEVICE = torch.device('cuda' if torch.cuda.is_available() else 'cpu')

    # Create save path directory
    if not os.path.exists(SAVE_PATH):
        os.makedirs(SAVE_PATH)

    # Get training data
    input_data = pd.read_csv(INPUT_DATA)
    output_data = pd.read_csv(OUTPUT_DATA)
    
    # Shuffle and split data
    x = input_data.to_numpy()
    y = output_data.to_numpy()
    x_scaler, x_std = get_std(x)
    y_scaler, y_std = get_std(y)
    x_train, x_valid, y_train, y_valid = train_test_split(x_std, y_std, test_size=0.2, shuffle=True, random_state=42)
    data_train, data_valid = [x_train, y_train], [x_valid, y_valid]

    # Model reset
    torch.manual_seed(42)
    model = torch.load(BEST_MODEL_PKL, weights_only=False)
    model.apply(init_weights)

    # Parse the lr and weight_decay from best_model_params.log
    with open(BEST_MODEL_PARAMS, 'r') as f:
        lines = f.readlines()
        for line in lines:
            if 'lr:' in line:
                LR = float(line.split('lr:')[1].strip())
            if 'weight_decay:' in line:
                WD = float(line.split('weight_decay:')[1].strip())

    # Retrain model
    model.train()
    optimizer = torch.optim.RMSprop(model.parameters(), lr=LR, weight_decay=WD)
    loss_func = torch.nn.MSELoss()
    best_loss = np.inf
    wait = 0
    train_loss = []
    valid_loss = []
    train_x, train_y = torch.FloatTensor(data_train[0]).to(DEVICE), torch.FloatTensor(data_train[1]).to(DEVICE)
    valid_x, valid_y = torch.FloatTensor(data_valid[0]).to(DEVICE), torch.FloatTensor(data_valid[1]).to(DEVICE)
    for epoch in tqdm(range(EPOCHS), desc='Training'):
        optimizer.zero_grad()
        output_train = model(train_x)
        output_valid = model(valid_x)
        loss_train = loss_func(output_train, train_y)
        loss_valid = loss_func(output_valid, valid_y)
        loss_train.backward()
        optimizer.step()
        train_loss.append([epoch, loss_train.item()])
        valid_loss.append([epoch, loss_valid.item()])

        # Early stopping
        if loss_valid.item() < best_loss:
            best_loss = loss_valid.item()
            wait = 0
        else:
            wait += 1
            if wait >= PATIENCE:
                break
    
    # Plot training loss vs. epoch and validation loss vs. epoch
    train_loss = np.array(train_loss)
    valid_loss = np.array(valid_loss)
    sns.set_theme(font_scale=1.2, style='whitegrid')
    matplotlib.rcParams['xtick.direction'] = 'in'
    matplotlib.rcParams['ytick.direction'] = 'in'
    fig = plt.figure(figsize=(10, 4), constrained_layout=True)
    ax = plt.subplot()
    ax.plot(train_loss[:, 0], train_loss[:, 1], label='Training', color='C0', linewidth=1.5)
    ax.plot(valid_loss[:, 0], valid_loss[:, 1], label='Validation', color='C1', linewidth=1.5)
    ax.set_ylabel('Loss')
    ax.set_xlabel('Epoch')
    ax.legend(loc='upper right', frameon=True)
    plt.savefig(f'{SAVE_PATH}/loss.png', dpi=330)
    plt.close()

    # Save model
    torch.save(model, SAVE_PATH + '/trained_model.pkl')

    # Evaluate model
    model.eval()
    with torch.no_grad():
        pred_y_train = model(train_x).cpu().numpy()
        pred_y_valid = model(valid_x).cpu().numpy()
    pred_y_train_rescaled = y_scaler.inverse_transform(pred_y_train)
    true_y_train_rescaled = y_scaler.inverse_transform(y_train)
    pred_y_valid_rescaled = y_scaler.inverse_transform(pred_y_valid)
    true_y_valid_rescaled = y_scaler.inverse_transform(y_valid)
    
    # Calculate RMSE and R2 on training and validation set
    rmse_train = root_mean_squared_error(true_y_train_rescaled, pred_y_train_rescaled)
    r2_train = r2_score(true_y_train_rescaled, pred_y_train_rescaled)
    rmse_valid = root_mean_squared_error(true_y_valid_rescaled, pred_y_valid_rescaled)
    r2_valid = r2_score(true_y_valid_rescaled, pred_y_valid_rescaled)
    with open(f'{SAVE_PATH}/performance.log', 'w') as f:
        f.write('Model Performance on Training Set:\n')
        f.write(f'RMSE: {rmse_train:.4f}\n')
        f.write(f'R2: {r2_train:.4f}\n\n')
        f.write('Model Performance on Validation Set:\n')
        f.write(f'RMSE: {rmse_valid:.4f}\n')
        f.write(f'R2: {r2_valid:.4f}\n')
    
    # For each output dimension plot predicted vs. true value
    n_outputs = output_data.shape[1]
    headers = output_data.columns.tolist()
    for i in range(n_outputs):
        # Callculate RMSE and R2 for this output
        rmse_train_i = root_mean_squared_error(true_y_train_rescaled[:, i], pred_y_train_rescaled[:, i])
        r2_train_i = r2_score(true_y_train_rescaled[:, i], pred_y_train_rescaled[:, i])
        rmse_valid_i = root_mean_squared_error(true_y_valid_rescaled[:, i], pred_y_valid_rescaled[:, i])
        r2_valid_i = r2_score(true_y_valid_rescaled[:, i], pred_y_valid_rescaled[:, i])

        sns.set_theme(font_scale=1.2, style='whitegrid')
        matplotlib.rcParams['xtick.direction'] = 'in'
        matplotlib.rcParams['ytick.direction'] = 'in'
        fig = plt.figure(figsize=(6, 6), constrained_layout=True)
        ax = plt.subplot()
        
        ax.scatter(true_y_train_rescaled[:, i], pred_y_train_rescaled[:, i], color='C0', label='Training', alpha=0.8, zorder=2)
        ax.scatter(true_y_valid_rescaled[:, i], pred_y_valid_rescaled[:, i], color='C1', label='Validation', alpha=0.8, zorder=2)
        min_val = min(true_y_train_rescaled[:, i].min(), pred_y_train_rescaled[:, i].min(),
                      true_y_valid_rescaled[:, i].min(), pred_y_valid_rescaled[:, i].min())
        max_val = max(true_y_train_rescaled[:, i].max(), pred_y_train_rescaled[:, i].max(),
                      true_y_valid_rescaled[:, i].max(), pred_y_valid_rescaled[:, i].max())
        ax.plot([min_val, max_val], [min_val, max_val], color='C3', linestyle='--', linewidth=1.0, zorder=1)
        ax.set_xlabel('True Value')
        ax.set_ylabel('Predicted Value')
        text = (
            f'Training:\n'
            f'    RMSE: {rmse_train_i:.4f}\n'
            f'    R2: {r2_train_i:.4f}\n\n'
            f'Validation:\n'
            f'    RMSE: {rmse_valid_i:.4f}\n'
            f'    R2: {r2_valid_i:.4f}'
        )
        ax.text(0.7, 0.05, 
                text,
                horizontalalignment='left', 
                verticalalignment='bottom', 
                transform=ax.transAxes,
                fontsize=12,
                bbox=dict(facecolor='white', alpha=0.8, edgecolor='gray')
                )
        ax.legend(loc='upper left', frameon=True)
        ax.set_title(f'Pred vs. True for {headers[i]}')
        plt.savefig(f'{SAVE_PATH}/pred_vs_true_{headers[i]}.png', dpi=330)
        plt.close()

if __name__ == '__main__':
    train(
        input_data=INPUT_DATA,
        output_data=OUTPUT_DATA,
        best_model_pkl=BEST_MODEL_PKL,
        best_model_params=BEST_MODEL_PARAMS,
        epochs=EPOCHS,
        patience=PATIENCE,
        save_path=SAVE_PATH,
    )


# !/usr/bin/env python3

import sys
from bullet import Bullet, colors

from masgent.cli_mode.cli_entries import run_command
from masgent.utils.utils import (
    color_print,
    print_help,
    start_new_session,
    clear_and_print_banner_and_entry_message,
    exit_and_cleanup,
    )

def main():
    # Create a single session runs directory
    start_new_session()
    
    try:
        while True:
            clear_and_print_banner_and_entry_message()
            choices = [
                '1. Density Functional Theory (DFT) Simulations',
                '2. Fast simulations using machine learning potentials (MLPs)',
                '3. Simple Machine Learning for Materials Science',
                '',
                'AI    ->  Chat with the Masgent AI',
                'New   ->  Start a new session',
                'Help  ->  Show available functions',
                'Exit  ->  Quit the Masgent',
            ]
            cli = Bullet(prompt='\n', choices=choices, margin=1, bullet=' ●', word_color=colors.foreground['green'])
            user_input = cli.launch()
            
            if user_input.startswith('AI'):
                from masgent.ai_mode import ai_backend
                ai_backend.main()
            if user_input.startswith('New'):
                start_new_session()
            elif user_input.startswith('Help'):
                print_help()
            elif user_input.startswith('Exit'):
                exit_and_cleanup()
            elif user_input.startswith('1'):
                run_command('1')
            elif user_input.startswith('2'):
                run_command('2')
            elif user_input.startswith('3'):
                run_command('3')
            else:
                continue
    
    except (KeyboardInterrupt, EOFError):
        exit_and_cleanup()

if __name__ == '__main__':
    main()



# !/usr/bin/env python3

import os, sys, datetime, time
import numpy as np
from pathlib import Path
from colorama import Fore, Style
from importlib.metadata import version, PackageNotFoundError

def create_deformation_matrices():
    xx = [-0.010, 0.010]
    yy = [-0.010, 0.010]
    zz = [-0.010, 0.010]
    xy = [-0.005, 0.005]
    yz = [-0.005, 0.005]
    xz = [-0.005, 0.005]
    D_000 = {f'00_strain_0.000': [
        [0.000, 0.000, 0.000], 
        [0.000, 0.000, 0.000], 
        [0.000, 0.000, 0.000]]}
    D_xx0 = {f'01_strain_xx_{float(xx[0]):.3f}': [
        [xx[0], 0.000, 0.000], 
        [0.000, 0.000, 0.000], 
        [0.000, 0.000, 0.000]]}
    D_yy0 = {f'03_strain_yy_{float(yy[0]):.3f}': [
        [0.000, 0.000, 0.000], 
        [0.000, yy[0], 0.000], 
        [0.000, 0.000, 0.000]]}
    D_zz0 = {f'05_strain_zz_{float(zz[0]):.3f}': [
        [0.000, 0.000, 0.000], 
        [0.000, 0.000, 0.000], 
        [0.000, 0.000, zz[0]]]}
    D_xy0 = {f'07_strain_xy_{float(xy[0]):.3f}': [
        [0.000, xy[0], 0.000], 
        [xy[0], 0.000, 0.000], 
        [0.000, 0.000, 0.000]]}
    D_yz0 = {f'09_strain_yz_{float(yz[0]):.3f}': [
        [0.000, 0.000, 0.000], 
        [0.000, 0.000, yz[0]], 
        [0.000, yz[0], 0.000]]}
    D_xz0 = {f'11_strain_xz_{float(xz[0]):.3f}': [
        [0.000, 0.000, xz[0]], 
        [0.000, 0.000, 0.000], 
        [xz[0], 0.000, 0.000]]}
    D_xx1 = {f'02_strain_xx_{float(xx[1]):.3f}': [
        [xx[1], 0.000, 0.000], 
        [0.000, 0.000, 0.000], 
        [0.000, 0.000, 0.000]]}
    D_yy1 = {f'04_strain_yy_{float(yy[1]):.3f}': [
        [0.000, 0.000, 0.000], 
        [0.000, yy[1], 0.000], 
        [0.000, 0.000, 0.000]]}
    D_zz1 = {f'06_strain_zz_{float(zz[1]):.3f}': [
        [0.000, 0.000, 0.000], 
        [0.000, 0.000, 0.000], 
        [0.000, 0.000, zz[1]]]}
    D_xy1 = {f'08_strain_xy_{float(xy[1]):.3f}': [
        [0.000, xy[1], 0.000], 
        [xy[1], 0.000, 0.000], 
        [0.000, 0.000, 0.000]]}
    D_yz1 = {f'10_strain_yz_{float(yz[1]):.3f}': [
        [0.000, 0.000, 0.000], 
        [0.000, 0.000, yz[1]], 
        [0.000, yz[1], 0.000]]}
    D_xz1 = {f'12_strain_xz_{float(xz[1]):.3f}': [
        [0.000, 0.000, xz[1]], 
        [0.000, 0.000, 0.000], 
        [xz[1], 0.000, 0.000]]}
    D_all = [D_000, D_xx0, D_yy0, D_zz0, D_xy0, D_yz0, D_xz0, D_xx1, D_yy1, D_zz1, D_xy1, D_yz1, D_xz1]
    return D_all

def eos_func(volume, a, b, c, d):
    energy = a + b * volume**(-2/3) + c * volume**(-4/3) + d * volume**(-2)
    return energy

def fit_eos(volumes, energies):
    from scipy.optimize import curve_fit

    volumes_fit = np.linspace(min(volumes) * 0.99, max(volumes) * 1.01, 100)
    popt, pcov = curve_fit(eos_func, volumes, energies)
    energies_fit = eos_func(volumes_fit, *popt)
    return volumes_fit, energies_fit

def list_files_in_dir(dir):
    '''List all files in a directory and its subdirectories.'''
    base_dir = os.environ.get('MASGENT_SESSION_RUNS_DIR')
    all_files = []
    for root, dirs, files in os.walk(dir):
        for file in files:
            all_files.append(os.path.join(base_dir, file))
    return all_files

def start_new_session():
    '''Set up a new session runs directory.'''
    base_dir = os.getcwd()
    main_dir = os.path.join(base_dir, 'masgent_projects')
    os.makedirs(main_dir, exist_ok=True)
    
    # Create a new runs directory with timestamp
    timestamp = datetime.datetime.now().strftime('%Y%m%d%H%M%S')
    runs_dir = os.path.join(main_dir, f'runs_{timestamp}')
    if not os.path.exists(runs_dir):
        os.makedirs(runs_dir, exist_ok=True)
    else:
        # Rare collision case, wait a second and try again
        time.sleep(1)
        timestamp = datetime.datetime.now().strftime('%Y%m%d%H%M%S')
        runs_dir = os.path.join(main_dir, f'runs_{timestamp}')
        os.makedirs(runs_dir, exist_ok=True)

    os.environ['MASGENT_SESSION_RUNS_DIR'] = runs_dir

def exit_and_cleanup():
    '''Exit Masgent and clean up empty runs directory.'''
    runs_dir = os.environ.get('MASGENT_SESSION_RUNS_DIR')
    if runs_dir and os.path.exists(runs_dir) and not os.listdir(runs_dir):
        os.rmdir(runs_dir)
    color_print('\nExiting Masgent... Goodbye!\n', 'green')
    sys.exit(0)

def global_commands():
    return [
        '',
        'AI    ->  Chat with the Masgent AI',
        'New   ->  Start a new session',
        'Back  ->  Return to previous menu',
        'Main  ->  Return to main menu',
        'Help  ->  Show available functions',
        'Exit  ->  Quit the Masgent',
    ]

def write_comments(file, file_type, comments):
    with open(file, 'r') as f:
        lines = f.readlines()

    if file_type.lower() in {'poscar', 'kpoints'}:
        lines[0] = comments + '\n'
    
    elif file_type.lower() in {'incar'}:
        lines.insert(0, f'{comments}\n')
    
    with open(file, 'w') as f:
        f.writelines(lines)

def generate_submit_script():
    '''
    Generate a basic Slurm submission script for VASP jobs.
    '''
    scripts = '''#!/bin/bash
#SBATCH --partition=normal
#SBATCH --nodes=1
#SBATCH --ntasks=8
#SBATCH --time=01:00:00
#SBATCH --job-name=masgent_job
#SBATCH --output=masgent_job.out
#SBATCH --error=masgent_job.err

# This Slurm script was generated by Masgent, customize as needed.

time srun vasp_std > vasp.out
'''
    return scripts

def generate_batch_script(update_incar=True, update_kpoints=True):
    '''
    Generate batch script for HPC job submission.
    '''
    script_lines = []
    script_lines.append('''#!/bin/bash

# This script was generated by Masgent to update VASP input files and submit jobs.
# After modifying the template files (INCAR_temp, KPOINTS_temp, POTCAR_temp,
# and submit_temp.sh, if present), run:
#   bash ./RUN_ME.sh

# Update VASP inputs in each folder
for d in */; do
    cp POTCAR_temp  "$d/POTCAR"
''')
    if update_incar:
        script_lines.append('''
    cp INCAR_temp  "$d/INCAR"
''')
    if update_kpoints:
        script_lines.append('''
    cp KPOINTS_temp  "$d/KPOINTS"
''')
    script_lines.append('''
done

# Update submit script and submit jobs in each folder
for d in */; do
    cp submit_temp.sh  "$d/submit.sh"
    chmod +x "$d/submit.sh"
    cd "$d"
    sbatch submit.sh
    cd ..
done
''')
    scripts = ''.join(script_lines)
    return scripts

def get_color_map():
    return {
        'red': Fore.RED,
        'green': Fore.GREEN,
        'yellow': Fore.YELLOW,
        'blue': Fore.BLUE,
        'magenta': Fore.MAGENTA,
        'cyan': Fore.CYAN,
        'white': Fore.WHITE,
    }

def color_print(text, color='cyan'):
    '''Print text in specified color.'''
    color_map = get_color_map()
    chosen_color = color_map.get(color.lower(), Fore.CYAN)
    print(chosen_color + text + Style.RESET_ALL)

def color_input(text, color='cyan'):
    '''Input prompt in specified color.'''
    color_map = get_color_map()
    chosen_color = color_map.get(color.lower(), Fore.CYAN)
    return input(chosen_color + text + Style.RESET_ALL)

def load_system_prompts():
    # src/masgent/ai_mode/system_prompt.txt
    prompts_path = Path(__file__).resolve().parent / 'ai_mode' / 'system_prompt.txt'
    try:
        return prompts_path.read_text(encoding='utf-8')
    except Exception as e:
        return f'Error loading system prompts: {str(e)}'

def validate_openai_api_key(key):
    try:
        from openai import OpenAI
        client = OpenAI(api_key=key)
        client.models.list()
        # color_print('[Info] OpenAI API key validated successfully.\n', 'green')
    except Exception as e:
        color_print('[Error] Invalid OpenAI API key. Exiting...\n', 'green')
        sys.exit(1)

def ask_for_openai_api_key():
    key = color_input('Enter your OpenAI API key: ', 'yellow').strip()
    if not key:
        color_print('[Error] OpenAI API key cannot be empty. Exiting...\n', 'green')
        sys.exit(1)
    
    validate_openai_api_key(key)

    os.environ['OPENAI_API_KEY'] = key

    save = color_input('Save this key to .env file for future? (y/n): ', 'yellow').strip().lower()
    base_dir = os.getcwd()
    env_path = os.path.join(base_dir, '.env')
    if save == 'y':
        with open(env_path, 'w') as f:
            f.write(f'OPENAI_API_KEY={key}\n')
        color_print(f'[Info] OpenAI API key saved to {env_path} file.\n', 'green')
    
def validate_mp_api_key(key):
    try:
        from mp_api.client import MPRester
        with MPRester(key, mute_progress_bars=True) as mpr:
            _ = mpr.materials.search(
                formula='Si',
                fields=['material_id']
            )
        # color_print('[Info] Materials Project API key validated successfully.\n', 'green')
    except Exception as e:
        color_print('[Error] Invalid Materials Project API key. Exiting...\n', 'green')
        sys.exit(1)
    
def ask_for_mp_api_key():
    key = color_input('Enter your Materials Project API key: ', 'yellow').strip()
    if not key:
        color_print('[Error] Materials Project API key cannot be empty. Exiting...\n', 'green')
        sys.exit(1)

    validate_mp_api_key(key)

    os.environ['MP_API_KEY'] = key

    save = color_input('Save this key to .env file for future? (y/n): ', 'yellow').strip().lower()
    base_dir = os.getcwd()
    env_path = os.path.join(base_dir, '.env')
    if save == 'y':
        with open(env_path, 'a') as f:
            f.write(f'MP_API_KEY={key}\n')
        color_print(f'[Info] Materials Project API key saved to {env_path} file.\n', 'green')

def print_banner():
    try:
        pkg_version = version('masgent')
    except PackageNotFoundError:
        pkg_version = 'dev'
    
    ascii_banner = rf'''
╔═════════════════════════════════════════════════════════════════════════╗
║                                                                         ║
║  ███╗   ███╗  █████╗  ███████╗  ██████╗  ███████╗ ███╗   ██╗ ████████╗  ║
║  ████╗ ████║ ██╔══██╗ ██╔════╝ ██╔════╝  ██╔════╝ ████╗  ██║ ╚══██╔══╝  ║
║  ██╔████╔██║ ███████║ ███████╗ ██║  ███╗ █████╗   ██╔██╗ ██║    ██║     ║
║  ██║╚██╔╝██║ ██╔══██║ ╚════██║ ██║   ██║ ██╔══╝   ██║╚██╗██║    ██║     ║
║  ██║ ╚═╝ ██║ ██║  ██║ ███████║ ╚██████╔╝ ███████╗ ██║ ╚████║    ██║     ║
║  ╚═╝     ╚═╝ ╚═╝  ╚═╝ ╚══════╝  ╚═════╝  ╚══════╝ ╚═╝  ╚═══╝    ╚═╝     ║
║                                                                         ║
║                                   MASGENT: Materials Simulation Agent   ║
║                                      Copyright (c) 2025 Guangchen Liu   ║
║                                                                         ║
║  Version:         {pkg_version:<52}  ║
║  Licensed:        MIT License                                           ║
║  Repository:      https://github.com/aguang5241/masgent                 ║
║  Citation:        Liu, G. et al. (2025), DOI:XX.XXXX/XXXXX              ║
║  Contact:         gliu4@wpi.edu                                         ║
║                                                                         ║
╚═════════════════════════════════════════════════════════════════════════╝
    '''
    color_print(ascii_banner, 'yellow')

def clear_and_print_entry_message():
    os.system('cls' if os.name == 'nt' else 'clear')
    msg = f'''
Welcome to Masgent AI — Your Materials Simulations Agent.
---------------------------------------------------------
Current Session Runs Directory: {os.environ["MASGENT_SESSION_RUNS_DIR"]}

Please select from the following options:
'''
    color_print(msg, 'white')

def clear_and_print_banner_and_entry_message():
    os.system('cls' if os.name == 'nt' else 'clear')
    print_banner()
    msg = f'''
Welcome to Masgent — Your Materials Simulation Agent.
---------------------------------------------------------
Current Session Runs Directory: {os.environ["MASGENT_SESSION_RUNS_DIR"]}

Please select from the following options:
'''
    color_print(msg, 'white')

def print_help():
    os.system('cls' if os.name == 'nt' else 'clear')

    content = '''
Masgent - Available Commands and Functions: 
-------------------------------------------
1. Density Functional Theory (DFT) Simulations
  1.1 Structure Preparation & Manipulation
    1.1.1 Generate POSCAR from chemical formula
    1.1.2 Convert POSCAR coordinates (Direct <-> Cartesian)
    1.1.3 Convert structure file formats (CIF, POSCAR, XYZ)
    1.1.4 Generate structures with defects (Vacancies, Interstitials, Substitutions)
    1.1.5 Generate supercells
    1.1.6 Generate Special Quasirandom Structures (SQS)
    1.1.7 Generate surface slabs
    1.1.8 Generate interface structures
  1.2 VASP Input File Preparation
    1.2.1 Prepare full VASP input files (INCAR, KPOINTS, POTCAR, POSCAR)
    1.2.2 Generate INCAR templates (relaxation, static, MD, etc.)
    1.2.3 Generate KPOINTS with specified accuracy
    1.2.4 Generate HPC job submission script
  1.3 Standard VASP Workflows Preparation
    1.3.1 Convergence testing (ENCUT, KPOINTS)
    1.3.2 Equation of State (EOS)
    1.3.3 Elastic constants calculations
    1.3.4 Ab-initio Molecular Dynamics (AIMD)
    1.3.5 Nudged Elastic Band (NEB) calculations
  1.4 (Planned) Workflow Output Analysis
2. Fast Simulations Using Machine Learning Potentials (MLPs)
  * Supported MLPs:
    2.1 SevenNet
    2.2 CHGNet
    2.3 Orb-v3
    2.4 MatSim
  * Implemented Simulations for all MLPs:
    - Single Point Energy Calculation
    - Equation of State (EOS) Calculation
    - Elastic Constants Calculation
    - Molecular Dynamics Simulation (NVT)
3. Simple Machine Learning for Materials Science
  3.1 (Planned) Data Preparation & Feature Engineering
    3.1.1 Feature analysis and visualization
    3.1.2 Dimensionality reduction (if too many features)
    3.1.3 Data augmentation (if limited data)
  3.2 (Planned) Model Design & Hyperparameter Tuning
  3.3 (Planned) Model Training & Evaluation
'''
    color_print(content, "green")

    try:
        while True:
            input = color_input('Type "back" to return: ', 'yellow').strip().lower()
            if input == 'back':
                return
    except KeyboardInterrupt:
        return




# !/usr/bin/env python3

import sys
from bullet import Bullet, colors

from masgent.cli_mode.cli_run import register, run_command
from masgent.utils.utils import (
    color_print, 
    print_help, 
    global_commands, 
    start_new_session,
    clear_and_print_entry_message,
    clear_and_print_banner_and_entry_message,
    exit_and_cleanup,
    )


###############################################
#                                             #
# Below are wrappers for main command entries #
#                                             #
###############################################

@register('0', 'Entry point for Masgent CLI.')
def command_0():
    try:
        while True:
            clear_and_print_banner_and_entry_message()
            choices = [
                '1. Density Functional Theory (DFT) Simulations',
                '2. Fast simulations using machine learning potentials (MLPs)',
                '3. Simple Machine Learning for Materials Science',
                '',
                'AI    ->  Chat with the Masgent AI',
                'New   ->  Start a new session',
                'Help  ->  Show available functions',
                'Exit  ->  Quit the Masgent',
            ]
            cli = Bullet(prompt='\n', choices=choices, margin=1, bullet=' ●', word_color=colors.foreground['green'])
            user_input = cli.launch()
            
            if user_input.startswith('AI'):
                from masgent.ai_mode import ai_backend
                ai_backend.main()
            elif user_input.startswith('New'):
                start_new_session()
            elif user_input.startswith('Help'):
                print_help()
            elif user_input.startswith('Exit'):
                exit_and_cleanup()
            elif user_input.startswith('1'):
                run_command('1')
            elif user_input.startswith('2'):
                run_command('2')
            elif user_input.startswith('3'):
                run_command('3')
            else:
                continue
    
    except (KeyboardInterrupt, EOFError):
        exit_and_cleanup()

@register('1', 'Density Functional Theory (DFT) Simulations.')
def command_1():
    try:
        while True:
            clear_and_print_entry_message()
            choices = [
                '1.1 Structure Preparation & Manipulation',
                '1.2 VASP Input File Preparation',
                '1.3 Standard VASP Workflow Preparation',
                '1.4 Standard VASP Workflow Output Analysis',
            ] + global_commands()
            cli = Bullet(prompt='\n', choices=choices, margin=1, bullet=' ●', word_color=colors.foreground['green'])
            user_input = cli.launch()

            if user_input.startswith('AI'):
                from masgent.ai_mode import ai_backend
                ai_backend.main()
            elif user_input.startswith('New'):
                start_new_session()
            elif user_input.startswith('Back'):
                return
            elif user_input.startswith('Main'):
                run_command('0')
            elif user_input.startswith('Help'):
                print_help()
            elif user_input.startswith('Exit'):
                exit_and_cleanup()
            elif user_input.startswith('1.1'):
                run_command('1.1')
            elif user_input.startswith('1.2'):
                run_command('1.2')
            elif user_input.startswith('1.3'):
                run_command('1.3')
            elif user_input.startswith('1.4'):
                run_command('1.4')
            else:
                continue

    except (KeyboardInterrupt, EOFError):
        exit_and_cleanup()


@register('1.1', 'Structure Preparation & Manipulation.')
def command_1_1():
    try:
        while True:
            clear_and_print_entry_message()
            choices = [
                '1.1.1 Generate POSCAR from chemical formula',
                '1.1.2 Convert POSCAR coordinates (Direct <-> Cartesian)',
                '1.1.3 Convert structure file formats (CIF, POSCAR, XYZ)',
                '1.1.4 Generate structures with defects (Vacancies, Substitutions, Interstitials)',
                '1.1.5 Generate supercells',
                '1.1.6 Generate Special Quasirandom Structures (SQS)',
                '1.1.7 Generate surface slabs',
                '1.1.8 Generate interface structures',
            ] + global_commands()
            cli = Bullet(prompt='\n', choices=choices, margin=1, bullet=' ●', word_color=colors.foreground['green'])
            user_input = cli.launch()

            if user_input.startswith('AI'):
                from masgent.ai_mode import ai_backend
                ai_backend.main()
            elif user_input.startswith('New'):
                start_new_session()
            elif user_input.startswith('Back'):
                return
            elif user_input.startswith('Main'):
                run_command('0')
            elif user_input.startswith('Help'):
                print_help()
            elif user_input.startswith('Exit'):
                exit_and_cleanup()
            elif user_input.startswith('1.1.1'):
                run_command('1.1.1')
            elif user_input.startswith('1.1.2'):
                run_command('1.1.2')
            elif user_input.startswith('1.1.3'):
                run_command('1.1.3')
            elif user_input.startswith('1.1.4'):
                run_command('1.1.4')
            elif user_input.startswith('1.1.5'):
                run_command('1.1.5')
            elif user_input.startswith('1.1.6'):
                run_command('1.1.6')
            elif user_input.startswith('1.1.7'):
                run_command('1.1.7')
            elif user_input.startswith('1.1.8'):
                run_command('1.1.8')
            else:
                continue

    except (KeyboardInterrupt, EOFError):
        exit_and_cleanup()

@register('1.2', 'VASP Input File Preparation')
def command_1_2():
    try:
        while True:
            clear_and_print_entry_message()
            choices = [
                '1.2.1 Prepare full VASP input files (INCAR, KPOINTS, POTCAR, POSCAR)',
                '1.2.2 Generate INCAR templates (relaxation, static, etc.)',
                '1.2.3 Gernerate KPOINTS with specified accuracy',
                '1.2.4 Generate HPC job submission script',
            ] + global_commands()
            cli = Bullet(prompt='\n', choices=choices, margin=1, bullet=' ●', word_color=colors.foreground['green'])
            user_input = cli.launch()

            if user_input.startswith('AI'):
                from masgent.ai_mode import ai_backend
                ai_backend.main()
            elif user_input.startswith('New'):
                start_new_session()
            elif user_input.startswith('Back'):
                return
            elif user_input.startswith('Main'):
                run_command('0')
            elif user_input.startswith('Help'):
                print_help()
            elif user_input.startswith('Exit'):
                exit_and_cleanup()
            elif user_input.startswith('1.2.1'):
                run_command('1.2.1')
            elif user_input.startswith('1.2.2'):
                run_command('1.2.2')
            elif user_input.startswith('1.2.3'):
                run_command('1.2.3')
            elif user_input.startswith('1.2.4'):
                run_command('1.2.4')
            else:
                continue

    except (KeyboardInterrupt, EOFError):
        exit_and_cleanup()

@register('1.3', 'Standard VASP Workflows.')
def command_1_3():
    try:
        while True:
            clear_and_print_entry_message()
            choices = [
                '1.3.1 Convergence testing (ENCUT, KPOINTS)',
                '1.3.2 Equation of State (EOS)',
                '1.3.3 Elastic constants calculations',
                '1.3.4 Ab-initio Molecular Dynamics (AIMD)',
                '1.3.5 Nudged Elastic Band (NEB) calculations',
            ] + global_commands()
            cli = Bullet(prompt='\n', choices=choices, margin=1, bullet=' ●', word_color=colors.foreground['green'])
            user_input = cli.launch()

            if user_input.startswith('AI'):
                from masgent.ai_mode import ai_backend
                ai_backend.main()
            elif user_input.startswith('New'):
                start_new_session()
            elif user_input.startswith('Back'):
                return
            elif user_input.startswith('Main'):
                run_command('0')
            elif user_input.startswith('Help'):
                print_help()
            elif user_input.startswith('Exit'):
                exit_and_cleanup()
            elif user_input.startswith('1.3.1'):
                run_command('1.3.1')
            elif user_input.startswith('1.3.2'):
                run_command('1.3.2')
            elif user_input.startswith('1.3.3'):
                run_command('1.3.3')
            elif user_input.startswith('1.3.4'):
                run_command('1.3.4')
            elif user_input.startswith('1.3.5'):
                run_command('1.3.5')
            else:
                continue

    except (KeyboardInterrupt, EOFError):
        exit_and_cleanup()

@register('1.4', 'VASP Output Analysis')
def command_1_4():
    try:
        while True:
            clear_and_print_entry_message()
            choices = [
                '1.4.1 Convergence test analysis',
                '1.4.2 Equation of State (EOS) analysis',
                '1.4.3 Elastic constants analysis',
                '1.4.4 Ab-initio Molecular Dynamics (AIMD) analysis',
                '1.4.5 Nudged Elastic Band (NEB) analysis',
            ] + global_commands()
            cli = Bullet(prompt='\n', choices=choices, margin=1, bullet=' ●', word_color=colors.foreground['green'])
            user_input = cli.launch()

            if user_input.startswith('AI'):
                from masgent.ai_mode import ai_backend
                ai_backend.main()
            elif user_input.startswith('New'):
                start_new_session()
            elif user_input.startswith('Back'):
                return
            elif user_input.startswith('Main'):
                run_command('0')
            elif user_input.startswith('Help'):
                print_help()
            elif user_input.startswith('Exit'):
                exit_and_cleanup()
            elif user_input.startswith('1.4.1'):
                run_command('1.4.1')
            elif user_input.startswith('1.4.2'):
                run_command('1.4.2')
            elif user_input.startswith('1.4.3'):
                run_command('1.4.3')
            elif user_input.startswith('1.4.4'):
                run_command('1.4.4')
            elif user_input.startswith('1.4.5'):
                run_command('1.4.5')
            else:
                continue

    except (KeyboardInterrupt, EOFError):
        exit_and_cleanup()

@register('2', 'Fast Simulations Using Machine Learning Potentials (MLPs).')
def command_2():
    try:
        while True:
            clear_and_print_entry_message()
            choices = [
                '2.1 SevenNet',
                '2.2 CHGNet',
                '2.3 Orb-v3',
                '2.4 MatSim',
            ] + global_commands()
            cli = Bullet(prompt='\n', choices=choices, margin=1, bullet=' ●', word_color=colors.foreground['green'])
            user_input = cli.launch()

            if user_input.startswith('AI'):
                from masgent.ai_mode import ai_backend
                ai_backend.main()
            elif user_input.startswith('New'):
                start_new_session()
            elif user_input.startswith('Back'):
                return
            elif user_input.startswith('Main'):
                run_command('0')
            elif user_input.startswith('Help'):
                print_help()
            elif user_input.startswith('Exit'):
                exit_and_cleanup()
            elif user_input.startswith('2.1'):
                run_command('2.1')
            elif user_input.startswith('2.2'):
                run_command('2.2')
            elif user_input.startswith('2.3'):
                run_command('2.3')
            elif user_input.startswith('2.4'):
                run_command('2.4')
            else:
                continue

    except (KeyboardInterrupt, EOFError):
        exit_and_cleanup()

@register('3', 'Simple Machine Learning for Materials Science.')
def command_3():
    try:
        while True:
            clear_and_print_entry_message()
            choices = [
                '3.1 Dataset Preparation & Visualization',
                '3.2 Model Design & Hyperparameter Tuning',
                '3.3 Model Training & Evaluation',
            ] + global_commands()
            cli = Bullet(prompt='\n', choices=choices, margin=1, bullet=' ●', word_color=colors.foreground['green'])
            user_input = cli.launch()

            if user_input.startswith('AI'):
                from masgent.ai_mode import ai_backend
                ai_backend.main()
            elif user_input.startswith('New'):
                start_new_session()
            elif user_input.startswith('Back'):
                return
            elif user_input.startswith('Main'):
                run_command('0')
            elif user_input.startswith('Help'):
                print_help()
            elif user_input.startswith('Exit'):
                exit_and_cleanup()
            elif user_input.startswith('3.1'):
                run_command('3.1')
            elif user_input.startswith('3.2'):
                run_command('3.2')
            elif user_input.startswith('3.3'):
                run_command('3.3')
            else:
                continue

    except (KeyboardInterrupt, EOFError):
        exit_and_cleanup()

@register('3.1', 'Data Preparation & Feature Analysis')
def command_3_1():
    try:
        while True:
            clear_and_print_entry_message()
            choices = [
                '3.1.1 Feature analysis and visualization',
                '3.1.2 Dimensionality reduction (if too many features)',
                '3.1.3 Data augmentation (if limited data)'
            ] + global_commands()
            cli = Bullet(prompt='\n', choices=choices, margin=1, bullet=' ●', word_color=colors.foreground['green'])
            user_input = cli.launch()

            if user_input.startswith('AI'):
                from masgent.ai_mode import ai_backend
                ai_backend.main()
            elif user_input.startswith('New'):
                start_new_session()
            elif user_input.startswith('Back'):
                return
            elif user_input.startswith('Main'):
                run_command('0')
            elif user_input.startswith('Help'):
                print_help()
            elif user_input.startswith('Exit'):
                exit_and_cleanup()
            elif user_input.startswith('3.1.1'):
                run_command('3.1.1')
            elif user_input.startswith('3.1.2'):
                run_command('3.1.2')
            elif user_input.startswith('3.1.3'):
                run_command('3.1.3')
            else:
                continue

    except (KeyboardInterrupt, EOFError):
        exit_and_cleanup()



import numpy as np
import pandas as pd
from tqdm import tqdm
from sklearn.discriminant_analysis import StandardScaler
import torch
import torch.nn as nn
from torch.nn.functional import mse_loss
from torch.utils.data import DataLoader, TensorDataset

# Conditional Variational Autoencoder (CVAE) for data augmentation
class CVAE(nn.Module):
    def __init__(self, x_dim, y_dim, latent_dim=8, hidden_dims=[64, 32, 16]):
        super().__init__()
        self.x_dim = x_dim
        self.y_dim = y_dim
        self.latent_dim = latent_dim

        # Encoder: takes both x and y as input
        encoder_layers = []
        input_dim = x_dim + y_dim
        for h in hidden_dims:
            encoder_layers += [nn.Linear(input_dim, h), nn.ReLU()]
            input_dim = h
        self.encoder = nn.Sequential(*encoder_layers)

        self.fc_mu = nn.Linear(input_dim, latent_dim)
        self.fc_logvar = nn.Linear(input_dim, latent_dim)

        # Decoder: takes latent + y
        decoder_layers = []
        input_dim = latent_dim + y_dim
        for h in reversed(hidden_dims):
            decoder_layers += [nn.Linear(input_dim, h), nn.ReLU()]
            input_dim = h
        decoder_layers += [nn.Linear(input_dim, x_dim)]
        self.decoder = nn.Sequential(*decoder_layers)

    def encode(self, x, y):
        xy = torch.cat([x, y], dim=1)
        h = self.encoder(xy)
        mu = self.fc_mu(h)
        logvar = self.fc_logvar(h)
        return mu, logvar

    def reparameterize(self, mu, logvar):
        std = torch.exp(0.5 * logvar)
        eps = torch.randn_like(std)
        return mu + eps * std

    def decode(self, z, y):
        zy = torch.cat([z, y], dim=1)
        x_recon = self.decoder(zy)
        return x_recon

    def forward(self, x, y):
        mu, logvar = self.encode(x, y)
        z = self.reparameterize(mu, logvar)
        x_recon = self.decode(z, y)
        return x_recon, mu, logvar

# CVAE Loss Function
def cvae_loss(x, x_recon, mu, logvar):
    recon = mse_loss(x_recon, x, reduction='mean')
    kl = -0.5 * torch.mean(1 + logvar - mu.pow(2) - logvar.exp())
    return recon + kl

# Training function for CVAE
def train_cvae(model, loader, epochs=500, lr=1e-3, patience=50):
    opt = torch.optim.Adam(model.parameters(), lr=lr)
    best_loss = np.inf
    wait = 0

    for epoch in tqdm(range(epochs), desc="Training CVAE"):
        for x_batch, y_batch in loader:
            opt.zero_grad()
            x_recon, mu, logvar = model(x_batch, y_batch)
            loss = cvae_loss(x_batch, x_recon, mu, logvar)
            loss.backward()
            opt.step()

        # Early stopping
        if loss.item() < best_loss:
            best_loss = loss.item()
            wait = 0
        else:
            wait += 1
            if wait > patience:
                break

    return model

# Function to generate conditional samples using trained CVAE
def generate_conditional_samples(model, y_cond, num_samples=1):
    model.eval()
    latent_dim = model.latent_dim
    y_tensor = torch.tensor(y_cond, dtype=torch.float32)

    x_aug_list = []
    for _ in range(num_samples):
        z = torch.randn((1, latent_dim))
        x_aug = model.decode(z, y_tensor).detach().numpy().flatten()
        x_aug_list.append(x_aug)

    return np.array(x_aug_list)

# Main function to run CVAE-based data augmentation
def run_cvae_augmentation(input_df, output_df, num_aug=1000):
    x_dim = input_df.shape[1]
    y_dim = output_df.shape[1]

    # Scaling
    x_scaler = StandardScaler()
    y_scaler = StandardScaler()

    X = x_scaler.fit_transform(input_df.values)
    Y = y_scaler.fit_transform(output_df.values)

    dataset = TensorDataset(
        torch.tensor(X, dtype=torch.float32),
        torch.tensor(Y, dtype=torch.float32)
    )
    loader = DataLoader(dataset, batch_size=64, shuffle=True)

    cvae = CVAE(x_dim=x_dim, y_dim=y_dim, latent_dim=8, hidden_dims=[64, 32, 16])

    # Train CVAE
    cvae = train_cvae(cvae, loader)

    # Choose random outputs to condition on
    Y_choices = Y[np.random.choice(len(Y), size=num_aug, replace=True)]

    # Generate synthetic X conditioned on Y
    X_aug_std = []
    for y_cond in Y_choices:
        sample = generate_conditional_samples(cvae, y_cond.reshape(1, -1), num_samples=1)
        X_aug_std.append(sample[0])

    X_aug_std = np.array(X_aug_std)

    # Inverse transform to physical units
    X_aug = x_scaler.inverse_transform(X_aug_std)
    Y_aug = y_scaler.inverse_transform(Y_choices)

    x_aug_df = pd.DataFrame(X_aug, columns=input_df.columns)
    y_aug_df = pd.DataFrame(Y_aug, columns=output_df.columns)

    return x_aug_df, y_aug_df